"""Mysql module tests."""
from pathlib import Path
from unittest import mock

import pytest

from spicerack import mysql
from spicerack.constants import PUPPET_CA_PATH


class TestMysql:
    """Mysql class tests."""

    def make_match_from_defaults(self, kwargs, match):
        """Removes the need to put 'obvious' expectations in match."""
        if "charset" not in match:
            match["charset"] = kwargs.get("charset", "utf8mb4")

        if "read_default_file" not in match:
            match["read_default_file"] = kwargs.get("read_default_file", str(Path("~/.my.cnf").expanduser()))
        if "read_default_group" not in match:
            match["read_default_group"] = kwargs.get("read_default_group", "client")

        if "ssl" not in match:
            match["ssl"] = kwargs.get("ssl", {"ca": PUPPET_CA_PATH})

    @pytest.mark.parametrize(
        "kwargs, match",
        (
            ({}, {}),
            ({"host": "db9999"}, {"host": "db9999"}),
            ({"charset": "ascii"}, {"charset": "ascii"}),
            ({"charset": ""}, {"charset": ""}),
            ({"read_default_file": "/my.cnf"}, {"read_default_file": "/my.cnf"}),
            (
                {"read_default_file": None},
                {"read_default_file": None, "read_default_group": None},
            ),
            (
                {"read_default_group": "client_test"},
                {"read_default_group": "client_test"},
            ),
            (
                {"host": "labsdb9999"},
                {"host": "labsdb9999", "read_default_group": "clientlabsdb"},
            ),
            ({"ssl": {}}, {"ssl": {}}),
            ({"ssl": {1: 3}}, {"ssl": {1: 3}}),
        ),
    )
    @mock.patch("spicerack.mysql.Connection", autospec=True)
    def test_connect(self, mocked_pymsql_connect, kwargs, match):
        """It should call pymysql with the correct parameters."""
        my = mysql.Mysql(dry_run=False)
        self.make_match_from_defaults(kwargs, match)
        with my.connect(**kwargs) as conn:
            assert conn == mocked_pymsql_connect.return_value
            mocked_pymsql_connect.assert_called_once_with(**match)
            # Ensure close is not called before context manager exits
            conn.close.assert_not_called()  # pylint: disable=maybe-no-member
        conn.close.assert_called_once_with()  # pylint: disable=maybe-no-member

    @pytest.mark.parametrize(
        "dry_run, read_only, transaction_ro",
        (
            (False, False, False),
            (False, True, True),
            (True, False, True),
            (True, True, True),
        ),
    )
    @mock.patch("spicerack.mysql.Connection", autospec=True)
    def test_connect_read_only(self, mocked_pymsql_connect, dry_run, read_only, transaction_ro):
        """It should start a read-only transaction if either dry-run or read-only are set."""
        my = mysql.Mysql(dry_run=dry_run)
        with my.connect(read_only=read_only) as conn:
            assert conn == mocked_pymsql_connect.return_value
            if transaction_ro:
                conn.query.assert_called_once_with(  # pylint: disable=maybe-no-member
                    "SET SESSION TRANSACTION READ ONLY"
                )
            else:
                conn.query.assert_not_called()  # pylint: disable=maybe-no-member



"""Alertmanager module."""
import logging
import re
from collections.abc import Iterator, Mapping, Sequence
from contextlib import contextmanager
from datetime import datetime, timedelta, timezone
from typing import Optional, Union

from cumin import NodeSet, nodeset_fromlist
from requests import Response
from requests.exceptions import RequestException
from wmflib.requests import DEFAULT_RETRY_STATUS_CODES, http_session

from spicerack.administrative import Reason
from spicerack.exceptions import SpicerackError
from spicerack.typing import TypeHosts

logger = logging.getLogger(__name__)
MatchersType = Sequence[dict[str, Union[str, int, float, bool]]]
PORT_REGEX: str = r"(\..+)?(:[0-9]+)?"
"""The regular expression used to match FQDNs and port numbers in the instance labels."""
ALERTMANAGER_URLS: tuple[str, str] = (
    "http://alertmanager-eqiad.wikimedia.org",
    "http://alertmanager-codfw.wikimedia.org",
)
"""All the alertmanager instances to contact."""


class Alertmanager:
    """Operate on Alertmanager via its APIs."""

    def __init__(
        self,
        *,
        dry_run: bool = True,
    ) -> None:
        """Initialize the instance.

        When using Alertmanager in high availability (cluster) make sure to pass all hosts in your cluster as
        `alertmanager_urls`.

        Arguments:
            dry_run: whether this is a DRY-RUN.

        """
        # Alertmanager API returns HTTP 500 (Internal Server Error) on some requests with a valid JSON response
        # For example when trying to delete a silence that doesn't exist or has already been deleted or is expired
        # Do not retry on 500 and accept it's first response.
        self._http_session = http_session(
            ".".join((self.__module__, self.__class__.__name__)),
            timeout=2,
            retry_codes=tuple(i for i in DEFAULT_RETRY_STATUS_CODES if i != 500),
        )
        self._alertmanager_urls = ALERTMANAGER_URLS
        self._dry_run = dry_run

    def _api_request(self, method: str, path: str, json: Optional[Mapping] = None) -> Response:
        """Perform an Alertmanager API request on multiple endpoints and return the requests response object.

        The request is performed on all configured alertmanager endpoints and returns at the first successful response.

        Arguments:
            method: the HTTP method to use for the request.
            path: the final API path to call, the base path is prefixed automatically.
            json: if present, the JSON payload to send in the request.

        Raises:
            spicerack.alertmanager.AlertmanagerError: if unable to perform the request on any alertmanager endpoint.

        """
        response = None
        for am_url in self._alertmanager_urls:
            url = f"{am_url}/api/v2/{path}"
            if self._dry_run and method.lower() not in ("head", "get"):
                logger.debug("Would have called %s %s", method.upper(), url)
                response = Response()
                response.status_code = 200
                return response

            try:
                response = self._http_session.request(method, url, json=json)
                response.raise_for_status()
                return response
            except RequestException as e:
                logger.error("Failed to %s to %s: %s", method.upper(), url, e)

        raise AlertmanagerError(f"Unable to {method.upper()} to any Alertmanager: {self._alertmanager_urls}", response)

    @contextmanager
    def downtimed(
        self,
        reason: Reason,
        *,
        matchers: MatchersType,
        duration: timedelta = timedelta(hours=4),
        remove_on_error: bool = False,
    ) -> Iterator[None]:
        """Context manager to perform actions while the matching alerts are downtimed on Alertmanager.

        Arguments:
            reason: the reason to set for the downtime on Alertmanager.
            matchers: the list of matchers to be applied to the downtime. The downtime will match alerts that match
                **all** the matchers provided, as they are ANDed by AlertManager.
            duration: the length of the downtime period.
            remove_on_error: should the downtime be removed even if an exception was raised.

        Yields:
            None: it just yields control to the caller once Alertmanager has received the downtime and deletes the
            downtime once getting back the control.

        """
        downtime_id = self.downtime(reason, matchers=matchers, duration=duration)
        try:  # pylint: disable=no-else-raise
            yield
        except BaseException:
            if remove_on_error:
                self.remove_downtime(downtime_id)
            raise
        else:
            self.remove_downtime(downtime_id)

    def downtime(self, reason: Reason, *, matchers: MatchersType, duration: timedelta = timedelta(hours=4)) -> str:
        """Issue a new downtime.

        Arguments:
            reason: the downtime reason.
            matchers: the list of matchers to be applied to the downtime. The downtime will match alerts that match
                **all** the matchers provided, as they are ANDed by AlertManager.
            duration: the length of the downtime period.

        Returns:
            The downtime ID.

        Raises:
            spicerack.alertmanager.AlertmanagerError: if none of the `alertmanager_urls` API returned a success or the
            parameters are invalid.

        """
        if not matchers:
            raise AlertmanagerError("No matchers provided.")

        # Swagger API format for startsAt/endsAt is 'date-time' which includes a timezone.
        start = datetime.utcnow().astimezone(tz=timezone.utc)
        end = start + duration
        payload = {
            "matchers": list(matchers),
            "startsAt": start.isoformat(),
            "endsAt": end.isoformat(),
            "comment": str(reason),
            "createdBy": reason.owner,
        }
        response = self._api_request("post", "silences", json=payload)
        if self._dry_run:  # Bail out earlier as the next statement would fail
            return ""

        silence = response.json()["silenceID"]
        logger.info("Created silence ID %s", silence)
        return silence

    def remove_downtime(self, downtime_id: str) -> None:
        """Remove a downtime.

        Arguments:
            downtime_id: the downtime ID to remove.

        Raises:
            spicerack.alertmanager.AlertmanagerError: if none of the `alertmanager_urls` API returned a success.

        """
        try:
            self._api_request("delete", f"silence/{downtime_id}")
            logger.info("Deleted silence ID %s", downtime_id)
        except AlertmanagerError as e:
            if (
                e.response is not None
                and e.response.status_code == 500
                and "silence" in e.response.json()
                and "already expired" in e.response.json()
            ):
                logger.warning("Silence ID %s has been already deleted or is expired", downtime_id)
            else:
                raise


class AlertmanagerHosts(Alertmanager):
    """Operate on Alertmanager for a list of hosts via its APIs."""

    def __init__(
        self,
        target_hosts: TypeHosts,
        *,
        verbatim_hosts: bool = False,
        dry_run: bool = True,
    ) -> None:
        """Initialize the instance.

        Arguments:
            target_hosts: the target hosts either as a NodeSet instance or a sequence of strings.
            verbatim_hosts: if :py:data:`True` use the hosts passed verbatim as is, if instead
                :py:data:`False`, the default, consider the given target hosts as FQDNs and extract their hostnames to
                be used in Alertmanager.
            dry_run: whether this is a DRY-RUN.

        Raises:
            spicerack.alertmanager.AlertmanagerError: if no target hosts are provided.

        """
        super().__init__(dry_run=dry_run)
        if not verbatim_hosts:
            target_hosts = [target_host.split(".")[0] for target_host in target_hosts]

        if isinstance(target_hosts, NodeSet):
            self._target_hosts = target_hosts
        else:
            self._target_hosts = nodeset_fromlist(target_hosts)

        if not self._target_hosts:
            raise AlertmanagerError("Got empty target hosts list.")

        self._verbatim_hosts = verbatim_hosts

    @contextmanager
    def downtimed(
        self,
        reason: Reason,
        *,
        matchers: MatchersType = (),
        duration: timedelta = timedelta(hours=4),
        remove_on_error: bool = False,
    ) -> Iterator[None]:
        """Context manager to perform actions while the hosts are downtimed on Alertmanager.

        Arguments:
            reason: the reason to set for the downtime on Alertmanager.
            matchers: an optional list of matchers to be applied to the downtime. They will be added to the matcher
                automatically generated to match the current instance ``target_hosts`` hosts. For this reason the
                provided matchers cannot be for the ``instance`` property. The downtime will match alerts that match
                **all** the matchers provided, as they are ANDed by AlertManager.
            duration: the length of the downtime period.
            remove_on_error: should the downtime be removed even if an exception was raised.

        Yields:
            None: it just yields control to the caller once Alertmanager has received the downtime and deletes the
            downtime once getting back the control.

        """
        # Keep this kinda duplicated method from the parent class for a few reasons:
        # * It needs a different default value for the matchers argument.
        # * It needs a different docstring for documenting the different behaviour of the matchers argument.
        # * Calling super() within a contextmanager is not that trivial and will de-facto require more code.
        downtime_id = self.downtime(reason, matchers=matchers, duration=duration)
        try:  # pylint: disable=no-else-raise
            yield
        except BaseException:
            if remove_on_error:
                self.remove_downtime(downtime_id)
            raise
        else:
            self.remove_downtime(downtime_id)

    def downtime(self, reason: Reason, *, matchers: MatchersType = (), duration: timedelta = timedelta(hours=4)) -> str:
        """Issue a new downtime for the given hosts.

        Arguments:
            reason: the downtime reason.
            matchers: an optional list of matchers to be applied to the downtime. They will be added to the matcher
                automatically generated to match the current instance ``target_hosts`` hosts. For this reason the
                provided matchers cannot be for the ``instance`` property. The downtime will match alerts that match
                **all** the matchers provided, as they are ANDed by AlertManager.
            duration: the length of the downtime period.

        Returns:
            str: the downtime ID.

        Raises:
            spicerack.alertmanager.AlertmanagerError: if none of the `alertmanager_urls` API returned a success or the
            parameters are invalid.

        """
        if any(item.get("name") == "instance" for item in matchers):
            raise AlertmanagerError("Matchers cannot target the instance property.")

        # If none of the hosts has the port embedded, put the port regex only once at the end
        group_port = all(":" not in host for host in self._target_hosts)
        group_port_regex = PORT_REGEX if group_port else ""
        target_hosts = []
        for host in sorted(self._target_hosts):
            if group_port or ":" in host:
                target_hosts.append(re.escape(host))
            else:
                target_hosts.append(fr"{re.escape(host)}{PORT_REGEX}")

        target_regex = "|".join(target_hosts)
        target_matchers = list(matchers)
        target_matchers.append({"name": "instance", "value": fr"^({target_regex}){group_port_regex}$", "isRegex": True})
        return super().downtime(reason, matchers=target_matchers, duration=duration)


class AlertmanagerError(SpicerackError):
    """Custom exception class for errors of this module."""

    def __init__(self, message: str, response: Optional[Response] = None) -> None:
        """Initializes an AlertmanagerError instance with the API response instance.

        Arguments:
            message: the actual exception message.
            response: the requests response object, if present.

        """
        super().__init__(message)
        self.response = response



"""Netbox module."""
import logging
import warnings
from typing import Union

import pynetbox
from wmflib.requests import http_session

from spicerack.exceptions import SpicerackError

MANAGEMENT_IFACE_NAME: str = "mgmt"
"""The interface name used in Netbox for the OOB network."""
SERVER_ROLE_SLUG: str = "server"
"""Netbox role to identify servers."""
logger = logging.getLogger(__name__)


class NetboxError(SpicerackError):
    """General errors raised by this module."""


class NetboxAPIError(NetboxError):
    """Usually a wrapper for pynetbox.RequestError, errors that occur when accessing the API."""


class NetboxHostNotFoundError(NetboxError):
    """Raised when a host is not found for an operation."""


class Netbox:
    """Class which wraps Netbox API operations."""

    def __init__(self, url: str, token: str, *, dry_run: bool = True):
        """Create Netbox instance.

        Arguments:
            url: The Netbox top level URL (with scheme and port if necessary).
            token: A Netbox API token.
            dry_run: set to False to cause writes to Netbox to occur.

        """
        self._api = pynetbox.api(url, token=token, threading=True)
        self._api.http_session = http_session(".".join((self.__module__, self.__class__.__name__)))
        self._dry_run = dry_run

    @property
    def api(self) -> pynetbox.api:
        """Get the pynetbox instance to interact directly with Netbox APIs.

        Caution:
            When feasible use higher level functionalities.

        """
        return self._api

    def _fetch_host(self, hostname: str) -> pynetbox.core.response.Record:
        """Fetch a host (dcim.devices) object.

        Arguments:
            hostname: the name of the host to fetch.

        Raises:
            spicerack.netbox.NetboxAPIError: on API error.
            spicerack.netbox.NetboxError: on parameter error.
            spicerack.netbox.NetboxHostNotFoundError: if the host is not found.

        """
        try:
            host = self._api.dcim.devices.get(name=hostname)
        except pynetbox.RequestError as ex:
            # excepts on other errors
            raise NetboxAPIError("Error retrieving Netbox host") from ex

        if host is None:
            raise NetboxHostNotFoundError
        return host

    def _fetch_virtual_machine(self, hostname: str) -> pynetbox.core.response.Record:
        """Fetch a virtual machine (virtualization.virtual_machine) object.

        Arguments:
            hostname: the name of the host to fetch.

        Raises:
            spicerack.netbox.NetboxAPIError: on API error.
            spicerack.netbox.NetboxError: on parameter error.
            spicerack.netbox.NetboxHostNotFoundError: if the host is not found.

        """
        try:
            host = self._api.virtualization.virtual_machines.get(name=hostname)
        except pynetbox.RequestError as ex:
            # excepts on other errors
            raise NetboxAPIError("Error retrieving Netbox VM") from ex

        if host is None:
            raise NetboxHostNotFoundError

        return host

    def put_host_status(self, hostname: str, status: str) -> None:
        """Set the device status.

        .. deprecated:: v0.0.50
            use :py:class:`spicerack.netbox.NetboxServer` instead.

        Note:
           This method does not operate on virtual machines since they are updated automatically from Ganeti into
           Netbox.

        Arguments:
            hostname: the name of the host to operate on.
            status: A status label or name.

        Raises:
            spicerack.netbox.NetboxAPIError: on API error.
            spicerack.netbox.NetboxError: on parameter error.

        """
        warnings.warn("Deprecated method, use spicearack.netbox_server() instead", DeprecationWarning)
        status = status.lower()
        host = self._fetch_host(hostname)
        oldstatus = host.status

        if self._dry_run:
            logger.info(
                "Skipping Netbox status update in DRY-RUN mode for host %s %s -> %s",
                hostname,
                oldstatus,
                status,
            )
            return

        host.status = status
        try:
            save_result = host.save()
        except pynetbox.RequestError as ex:
            raise NetboxAPIError(f"Failed to save Netbox status for host {hostname} {oldstatus} -> {status}") from ex

        if save_result:
            logger.info(
                "Netbox status updated for host %s %s -> %s",
                hostname,
                oldstatus,
                status,
            )
        else:
            raise NetboxAPIError(f"Failed to update Netbox status for host {hostname} {oldstatus} -> {status}")

    def fetch_host_status(self, hostname: str) -> str:
        """Return the current status of a host as a string.

        .. deprecated:: v0.0.50
            use :py:class:`spicerack.netbox.NetboxServer` instead.

        Arguments:
            hostname: the name of the host status

        Raises:
            spicerack.netbox.NetboxAPIError: on API error.
            spicerack.netbox.NetboxError: on parameter error.
            spicerack.netbox.NetboxHostNotFoundError: if the host is not found.

        """
        warnings.warn("Deprecated method, use spicearack.netbox_server() instead", DeprecationWarning)
        try:
            return str(self._fetch_host(hostname).status)
        except NetboxHostNotFoundError:
            return str(self._fetch_virtual_machine(hostname).status)

    def fetch_host_detail(self, hostname: str) -> dict:
        """Return a dict containing details about the host.

        .. deprecated:: v0.0.50
            use :py:class:`spicerack.netbox.NetboxServer` instead.

        Arguments:
            hostname: the name of the host to retrieve.

        Raises:
            spicerack.netbox.NetboxAPIError: on API error.
            spicerack.netbox.NetboxError: on parameter error.
            spicerack.netbox.NetboxHostNotFoundError: if the host is not found.

        """
        warnings.warn("Deprecated method, use spicearack.netbox_server() instead", DeprecationWarning)
        is_virtual = False
        vm_cluster = "N/A"
        try:
            host = self._fetch_host(hostname)
        except NetboxHostNotFoundError:
            host = self._fetch_virtual_machine(hostname)
            is_virtual = True
            vm_cluster = host.cluster.name

        ret = host.serialize()
        ret["is_virtual"] = is_virtual
        ret["ganeti_cluster"] = vm_cluster
        return ret

    def get_server(self, hostname: str) -> "NetboxServer":
        """Return a NetboxServer instance for the given hostname.

        Arguments:
            hostname: the device hostname.

        Raises:
            spicerack.netbox.NetboxHostNotFoundError: if the device can't be found among physical or virtual devices.
            spicerack.netbox.NetboxError: if the device is not a server.

        """
        try:
            server = self._fetch_host(hostname)
        except NetboxHostNotFoundError:
            server = self._fetch_virtual_machine(hostname)

        return NetboxServer(api=self._api, server=server, dry_run=self._dry_run)


class NetboxServer:
    """Represent a Netbox device of role server or a virtual machine."""

    allowed_status_transitions: dict[str, tuple[str, ...]] = {
        "spare": ("planned", "failed", "decommissioned"),
        "planned": ("active", "failed", "decommissioned"),
        "failed": ("spare", "planned", "active", "decommissioned"),
        "active": ("failed", "decommissioned"),
        "decommissioned": ("planned", "spare"),
    }
    """Allowed transition between Netbox statuses.
    See https://wikitech.wikimedia.org/wiki/Server_Lifecycle#/media/File:Server_Lifecycle_Statuses.png"""

    def __init__(
        self,
        *,
        api: pynetbox.api,
        server: Union[pynetbox.models.dcim.Devices, pynetbox.models.virtualization.VirtualMachines],
        dry_run: bool = True,
    ):
        """Initialize the instance.

        Arguments:
            api: the API instance to connect to Netbox.
            server: the server object.

        Raises:
            spicerack.netbox.NetboxError: if the device is not of type server.

        """
        self._server = server
        self._api = api
        self._dry_run = dry_run
        self._cached_mgmt_fqdn = ""  # Cache the management interface as it would require an API call each time

        role = server.role.slug if self.virtual else server.device_role.slug
        if role != SERVER_ROLE_SLUG:
            raise NetboxError(f"Object of type {type(server)} has invalid role {role}, only server is allowed")

    @property
    def virtual(self) -> bool:
        """Getter to check if the server is physical or virtual.

        Returns:
            :py:data:`True` if the server is virtual, :py:data:`False` if physical.

        """
        return not hasattr(self._server, "rack")

    @property
    def status(self) -> str:
        """Get and set the server status.

        Modifying its value can be done only on physical devices and only between allowed transitions.

        The allowed transitions are defined in :py:data:`spicerack.netbox.Netbox.allowed_status_transitions`.

        Arguments:
            value: the name of the status to be set. It will be lower cased automatically.

        Raises:
            spicerack.netbox.NetboxError: if trying to set it on a virtual device or the status transision is not
                allowed.

        """
        return self._server.status.value

    @status.setter
    def status(self, value: str) -> None:
        """Set the device status. See the getter docstring for info.

        Arguments:
            value: the name of the status to be set. It will be lower cased automatically.

        Raises:
            spicerack.netbox.NetboxError: if used on a virtual device or the status transision is not allowed.

        """
        if self.virtual:
            raise NetboxError(
                f"Server {self._server.name} is a virtual machine, its Netbox status is automatically synced from "
                f"Ganeti."
            )

        current = self._server.status.value
        new = value.lower()
        allowed_transitions = NetboxServer.allowed_status_transitions.get(current, ())
        if new not in allowed_transitions:
            raise NetboxError(
                f"Forbidden Netbox status transition between {current} and {new} for device {self._server.name}. "
                f"Possible values are: {allowed_transitions}"
            )

        if self._dry_run:
            logger.info(
                "Skipping Netbox status change from %s to %s for device %s in DRY-RUN.", current, new, self._server.name
            )
            return

        self._server.status = value
        self._server.save()
        logger.debug("Updated Netbox status from %s to %s for device %s", current, new, self._server.name)

    @property
    def fqdn(self) -> str:
        """Return the FQDN of the device.

        Raises:
            spicerack.netbox.NetboxError: if the server has no FQDN defined in Netbox.

        """
        # Until https://phabricator.wikimedia.org/T253173 is fixed we can't use the primary_ip attribute
        for attr_name in ("primary_ip4", "primary_ip6"):
            address = getattr(self._server, attr_name)
            if address is not None and address.dns_name:
                return address.dns_name

        raise NetboxError(f"Server {self._server.name} does not have any primary IP with a DNS name set.")

    @property
    def mgmt_fqdn(self) -> str:
        """Return the management FQDN of the device.

        Raises:
            spicerack.netbox.NetboxError: for virtual servers or the server has no management FQDN defined in Netbox.

        """
        if self.virtual:
            raise NetboxError(f"Server {self._server.name} is a virtual machine, does not have a management address.")

        if self._cached_mgmt_fqdn:
            return self._cached_mgmt_fqdn

        address = self._api.ipam.ip_addresses.get(device=self._server.name, interface=MANAGEMENT_IFACE_NAME)
        # TODO: check also that address.assigned_object.mgmt_only is True if it will not generate anymore an additional
        #       API call to Netbox or the Netbox API become more efficient.
        if address is not None and address.dns_name:
            self._cached_mgmt_fqdn = address.dns_name
            return self._cached_mgmt_fqdn

        raise NetboxError(f"Server {self._server.name} has no management interface with a DNS name set.")

    @property
    def asset_tag_fqdn(self) -> str:
        """Return the management FQDN for the asset tag of the device.

        Raises:
            spicerack.netbox.NetboxError: for virtual servers or the server has no management FQDN defined in Netbox.

        """
        parts = self.mgmt_fqdn.split(".")
        parts[0] = self._server.asset_tag.lower()
        return ".".join(parts)

    def as_dict(self) -> dict:
        """Return a dict containing details about the server."""
        ret = dict(self._server)
        ret["is_virtual"] = self.virtual

        return ret



"""Cookbook internal module."""
import argparse
from typing import Optional

from spicerack import Spicerack, cookbook


class CookbookModuleRunnerBase(cookbook.CookbookRunnerBase):
    """Abstract class for the dynamically converted module based Cookbooks to class based."""

    def __init__(self, args: argparse.Namespace, spicerack: Spicerack):
        """Constructor, saves the parameters to be passed to the _run() method."""
        self.args = args
        self.spicerack = spicerack

    def run(self) -> Optional[int]:
        """As required by the parent class."""
        return self._run(self.args, self.spicerack)

    @staticmethod
    def _run(args: argparse.Namespace, spicerack: Spicerack) -> Optional[int]:
        """To be dynamically overwritten with the run() module function."""


class CookbooksModuleInterface:
    """Module interface to be used as type hint for the imported cookbooks that use the module API."""

    __name__ = ""
    """The module name."""

    __title__: str = ""
    """The cookbook static title."""

    @staticmethod
    def argument_parser() -> argparse.ArgumentParser:  # type: ignore[empty-body]
        """Optional module function to define if the cookbook should accept command line arguments."""

    @staticmethod
    def run(args: argparse.Namespace, spicerack: Spicerack) -> Optional[int]:
        """Mandatory module function that every cookbook using this interface must define."""



"""Confctl module to abstract Conftool."""
import logging
import re
from collections.abc import Iterable, Iterator
from contextlib import contextmanager
from typing import Union

from conftool import configuration, kvobject, loader
from conftool.drivers import BackendError

from spicerack.exceptions import SpicerackError

logger = logging.getLogger(__name__)


class ConfctlError(SpicerackError):
    """Custom exception class for errors of this module."""


class Confctl:
    """Confctl class to abstract conftool operations."""

    def __init__(
        self,
        config: str = "/etc/conftool/config.yaml",
        schema: str = "/etc/conftool/schema.yaml",
        dry_run: bool = True,
    ) -> None:
        """Initialize the instance.

        Arguments:
            config: the path to the configuration file to load.
            schema: the path to the Conftool schema to load.
            dry_run: whether this is a DRY-RUN.

        """
        self._dry_run = dry_run
        self._schema = loader.Schema.from_file(schema)
        kvobject.KVObject.setup(configuration.get(config))

    def entity(self, entity_name: str) -> "ConftoolEntity":
        """Get the Conftool specific entity class.

        Arguments:
            entity_name: the name of the entity..

        """
        return ConftoolEntity(self._schema.entities[entity_name], dry_run=self._dry_run)


class ConftoolEntity:
    """ConftoolEntity class to perform operations on a specific Conftool entity."""

    def __init__(self, entity: kvobject.Entity, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            entity: an instance of Conftool entity.
            dry_run: whether this is a DRY-RUN.

        """
        self._entity = entity
        self._dry_run = dry_run

    def _select(self, tags: dict[str, str]) -> Iterator[kvobject.Entity]:
        """Generator that yields the selected objects based on the provided tags.

        Arguments:
            tags: dictionary with tag: expression pairs of Conftool selectors.

        Yields:
            conftool.kvobject.Entity: the selected object.

        Raises:
            spicerack.confctl.ConfctlError: if no match is found.

        """
        selectors = {}
        for tag, expr in tags.items():
            selectors[tag] = re.compile(f"^{expr}$")

        obj = None
        for obj in self._entity.query(selectors):
            yield obj

        if obj is None:
            raise ConfctlError("No match found")

    def update(self, changed: dict[str, Union[bool, str, int, float]], **tags: str) -> None:
        """Updates the value of conftool objects corresponding to the selection done with tags.

        Arguments:
            changed: the new values to set for the selected objects.
            **tags: arbitrary Conftool tags as keyword arguments.

        Raises:
            spicerack.confctl.ConfctlError: on etcd or Conftool errors.

        Examples:
            >>> confctl.update({'pooled': 'no'}, service='appservers-.*', name='eqiad')

        """
        logger.debug("Updating conftool matching tags: %s", tags)
        self.update_objects(changed, self._select(tags))

    def get(self, **tags: str) -> Iterator[kvobject.Entity]:
        """Generator that yields conftool objects corresponding to the selection.

        Arguments:
            **tags: arbitrary Conftool tags as keyword arguments.

        Yields:
            conftool.kvobject.Entity: the selected object.

        """
        for obj in self._select(tags):
            logger.debug("Selected conftool object: %s", obj)
            yield obj

    def set_and_verify(self, key: str, value: Union[bool, str, int, float], **tags: str) -> None:
        """Set and verify a single Conftool value.

        Arguments:
            key: the key in Conftool to modify.
            value: the value to set.
            **tags: arbitrary Conftool tags as keyword arguments.

        Raises:
            spicerack.confctl.ConfctlError: on etcd or Conftool errors or failing to verify the changes.

        """
        logger.info("Setting %s=%s for tags: %s", key, value, tags)
        self.update({key: value}, **tags)

        for obj in self.get(**tags):  # Verify the changes were applied
            new = getattr(obj, key)
            if new != value and not self._dry_run:
                raise ConfctlError(f"Conftool key {key} has value '{new}', expecting '{value}' for tags: {tags}")

    def filter_objects(
        self, filter_expr: dict[str, Union[bool, str, int, float]], **tags: str
    ) -> Iterator[kvobject.Entity]:
        """Filters objects coming from conftool based on values.

        A generator will be returned which will contain only objects that match all filters.

        Arguments:
           filter_expr: a set of desired field names and values.
           **tags: arbitrary Conftool tags as keyword arguments.

        Yields:
            conftool.kvobject.Entity: the selected object.

        Raises:
            spicerack.confctl.ConfctlError: if no object corresponds to the tags.

        """
        for obj in self._select(tags):
            matching = True
            for key, desired in filter_expr.items():
                try:
                    value = getattr(obj, key)
                except AttributeError as e:
                    raise ConfctlError(f'Could not find property "{key}" in object {obj.pprint()}') from e
                if value != desired:
                    matching = False
            if matching:
                yield obj

    def update_objects(
        self,
        changed: dict[str, Union[bool, str, int, float]],
        objects: Iterable[kvobject.Entity],
    ) -> None:
        """Updates the value of the provided conftool objects.

        Examples:
            >>> inactive = confctl.filter_objects({'pooled': 'inactive'}, service='appservers-.*', name='eqiad')
            >>> confctl.update_objects({'pooled': 'no'}, inactive)

        Arguments:
            changed: the new values to set for the selected objects.
            query: an iterator of conftool objects.

        Raises:
            spicerack.confctl.ConfctlError: on etcd or Conftool errors.

        """
        # TODO: make the api nicer by returning an EntitiesCollection from filter_objects so we can allow to write
        # >>> inactive.update({'pooled': 'no'})
        if self._dry_run:
            message_prefix = "Skipping conftool update on dry-run mode"
        else:
            message_prefix = "Updating conftool"

        for obj in objects:
            logger.debug("%s: %s -> %s", message_prefix, obj, changed)
            if self._dry_run:
                continue

            try:
                obj.update(changed)
            except BackendError as e:
                raise ConfctlError("Error writing to etcd") from e
            except Exception as e:
                raise ConfctlError("Generic error in conftool") from e

    @contextmanager
    def change_and_revert(
        self, field: str, original: Union[bool, str, int, float], changed: Union[bool, str, int, float], **tags: str
    ) -> Iterator[Iterable[kvobject.Entity]]:
        """Context manager to perform actions with a changed value in conftool.

        This method will only act on objects that had the original value.

        Warning:
            If the code executed within the contextmanager raises an unhandled
            exception, the original state of the objects will NOT be restored.

        Arguments:
            field: The field to change the value of.
            original: the original value.
            changed: changed value.
            tags: Appropriate conftool tags for the chosen entity to select objects.

        Yields:
            generator: conftool.kvObject.Entity the objects that were acted upon.

        """
        objects = list(self.filter_objects({field: original}, **tags))
        self.update_objects({field: changed}, objects)
        yield objects
        self.update_objects({field: original}, objects)



"""Spicerack package."""
from collections.abc import Callable, Sequence
from ipaddress import ip_interface
from logging import Logger, getLogger
from pathlib import Path
from socket import gethostname
from typing import TYPE_CHECKING, Any, Optional

from git import Repo
from pkg_resources import DistributionNotFound, get_distribution
from wmflib import requests
from wmflib.actions import ActionsDict
from wmflib.config import load_ini_config, load_yaml_config
from wmflib.dns import Dns
from wmflib.interactive import get_username
from wmflib.phabricator import Phabricator, create_phabricator
from wmflib.prometheus import Prometheus, Thanos

from spicerack._log import irc_logger, sal_logger
from spicerack.administrative import Reason
from spicerack.alerting import AlertingHosts
from spicerack.alertmanager import Alertmanager, AlertmanagerHosts
from spicerack.apt import AptGetHosts
from spicerack.confctl import Confctl, ConftoolEntity
from spicerack.debmonitor import Debmonitor
from spicerack.dhcp import DHCP
from spicerack.dnsdisc import Discovery
from spicerack.elasticsearch_cluster import ElasticsearchClusters, create_elasticsearch_clusters
from spicerack.exceptions import SpicerackError
from spicerack.ganeti import Ganeti
from spicerack.icinga import ICINGA_DOMAIN, IcingaHosts
from spicerack.interactive import get_management_password
from spicerack.ipmi import Ipmi
from spicerack.k8s import Kubernetes
from spicerack.kafka import Kafka
from spicerack.mediawiki import MediaWiki
from spicerack.mysql import Mysql
from spicerack.mysql_legacy import MysqlLegacy
from spicerack.netbox import MANAGEMENT_IFACE_NAME, Netbox, NetboxServer
from spicerack.peeringdb import PeeringDB
from spicerack.puppet import PuppetHosts, PuppetMaster, PuppetServer, get_puppet_ca_hostname
from spicerack.redfish import Redfish, RedfishDell
from spicerack.redis_cluster import RedisCluster
from spicerack.remote import Remote, RemoteHosts
from spicerack.reposync import RepoSync
from spicerack.service import Catalog
from spicerack.toolforge.etcdctl import EtcdctlController
from spicerack.typing import TypeHosts

if TYPE_CHECKING:  # Imported only during type checking, prevents cyclic imports at runtime
    from spicerack._menu import BaseItem  # pragma: no cover


logger = getLogger(__name__)

try:
    __version__: str = get_distribution("wikimedia-spicerack").version  # Must be the same used as 'name' in setup.py
    """The version of the current Spicerack module."""
except DistributionNotFound:  # pragma: no cover - this should never happen during tests
    pass  # package is not installed


class Spicerack:  # pylint: disable=too-many-instance-attributes
    """Spicerack service locator."""

    def __init__(
        self,
        *,
        verbose: bool = False,
        dry_run: bool = True,
        cumin_config: str = "/etc/cumin/config.yaml",
        cumin_installer_config: str = "/etc/cumin/config-installer.yaml",
        conftool_config: str = "/etc/conftool/config.yaml",
        conftool_schema: str = "/etc/conftool/schema.yaml",
        debmonitor_config: str = "/etc/debmonitor.conf",
        spicerack_config_dir: str = "/etc/spicerack",
        http_proxy: str = "",
        get_cookbook_callback: Optional[Callable[["Spicerack", str, Sequence[str]], Optional["BaseItem"]]] = None,
        extender_class: Optional[type["SpicerackExtenderBase"]] = None,
    ) -> None:
        """Initialize the service locator for the Spicerack library.

        Arguments:
            verbose: whether to set the verbose mode.
            dry_run: whether this is a DRY-RUN.
            cumin_config: the path to Cumin's configuration file.
            cumin_installer_config: the path to Cumin's configuration file for the Debian installer.
            conftool_config: the path to Conftool's configuration file.
            conftool_schema: the path to Conftool's schema file.
            debmonitor_config: the path to Debmonitor's INI configuration file. It must have at least the following
                schema::

                    [DEFAULT]
                    server=debmonitor.example.com
                    cert=/etc/debmonitor/ssl/cert.pem
                    key=/etc/debmonitor/ssl/server.key

            spicerack_config_dir: the path for the root configuration directory for Spicerack. Module-specific
                configuration will be loaded from ``config_dir/module_name/``.
            http_proxy: the ``scheme://url:port`` of the HTTP proxy to use for external calls.
            get_cookbook_callback: a callable to retrieve a CookbookItem to execute a cookbook from inside another
                cookbook.
            extender_class: an optional class object that inherits from :py:class:`spicerack.SpicerackExtenderBase` to
                dynamically add accessors to Spicerack. If not set no extenders will be registered, even if
                ``external_modules_dir`` is specified in the configuration.

        """
        # Attributes
        self._verbose = verbose
        self._dry_run = dry_run
        self._http_proxy = http_proxy
        self._cumin_config = cumin_config
        self._cumin_installer_config = cumin_installer_config
        self._conftool_config = conftool_config
        self._conftool_schema = conftool_schema
        self._debmonitor_config = debmonitor_config
        self._spicerack_config_dir = Path(spicerack_config_dir)
        self._get_cookbook_callback = get_cookbook_callback

        self._username = get_username()
        self._current_hostname = gethostname()
        self._irc_logger = irc_logger
        self._sal_logger = sal_logger
        self._confctl: Optional[Confctl] = None
        self._service_catalog: Optional[Catalog] = None
        self._management_password: str = ""
        self._actions = ActionsDict()
        self._authdns_servers: dict[str, str] = {}

        self._extender = None
        if extender_class is not None:  # If present, instantiate it with the current instance as parameter
            self._extender = extender_class(spicerack=self)

    def __getattr__(self, name: str) -> Any:
        """Attribute accessor to dynamically load external accessors when present.

        This method is called only if a method or attribute with the given name is not present in the current instance.

        :Parameters:
            according to Python's Data model :py:meth:`object.__getattr__`.

        """
        if self._extender is not None:
            return getattr(self._extender, name)

        raise AttributeError(f"AttributeError: '{self.__class__.__name__}' object has no attribute '{name}'")

    @property
    def dry_run(self) -> bool:
        """Returns :py:data:`True` if the ``DRY-RUN`` mode is set, :py:data:`False` otherwise."""
        return self._dry_run

    @property
    def verbose(self) -> bool:
        """Returns :py:data:`True` if the verbose mode is set, :py:data:`False` otherwise."""
        return self._verbose

    @property
    def username(self) -> str:
        """Returns the name of the effective running user."""
        return self._username

    @property
    def config_dir(self) -> Path:
        """Getter for Spicerack's configuration file directory where the module-specific configs are available.

        They can be accessed with::

            self.config_dir / module_name / config_file

        """
        return self._spicerack_config_dir

    @property
    def http_proxy(self) -> str:
        """Returns the ``scheme://url:port`` of the HTTP PROXY proxy."""
        return self._http_proxy

    @property
    def requests_proxies(self) -> Optional[dict[str, str]]:
        """Returns the HTTP proxy configuration for the Requests module or :py:data:`None` if no HTTP proxy is set.

        See Also:
            https://requests.readthedocs.io/en/master/user/advanced/#proxies

        """
        if not self._http_proxy:
            return None

        return {"http": self._http_proxy, "https": self._http_proxy}

    @property
    def irc_logger(self) -> Logger:
        """Returns the logger instance to write to IRC in the #wikimedia-operations channel."""
        return self._irc_logger

    @property
    def sal_logger(self) -> Logger:
        """Returns the logger instance to write to IRC in the #wikimedia-operations and logging to SAL."""
        return self._sal_logger

    @property
    def actions(self) -> ActionsDict:
        """Returns a dictionary to log and record cookbook actions."""
        return self._actions

    @property
    def icinga_master_host(self) -> RemoteHosts:
        """Returns the instance to execute commands on the Icinga master host."""
        return self.remote().query(self.dns().resolve_cname(ICINGA_DOMAIN))

    @property
    def netbox_master_host(self) -> RemoteHosts:
        """Returns the instance to execute commands on the Netbox master host."""
        dns = self.dns()
        netbox_hostname = dns.resolve_ptr(dns.resolve_ips("netbox.discovery.wmnet")[0])[0]
        return self.remote().query(netbox_hostname)

    @property
    def management_password(self) -> str:
        """Returns the management password.

        It will be asked to the user if not already cached by the current instance.

        """
        if not self._management_password:
            self._management_password = get_management_password()

        return self._management_password

    @property
    def authdns_servers(self) -> dict[str, str]:
        """Getter for the authoritative DNS nameservers currently active in production.

        Returns a dictionary where keys are the hostnames and values are the IPs of the active authoritative
        nameservers.
        """
        if not self._authdns_servers:
            self._authdns_servers = load_yaml_config(self._spicerack_config_dir / "discovery" / "authdns.yaml")

        return self._authdns_servers

    @property
    def authdns_active_hosts(self) -> RemoteHosts:
        """Get a RemoteHosts instance to target the active authoritative nameservers.

        Examples:
            ::

                >>> authdns_hosts = spicerack.authdns_active_hosts
                >>> authdns_hosts.run_sync('some command')

        """
        hosts = ",".join(self.authdns_servers.keys())
        return self.remote().query(f"D{{{hosts}}}")

    def run_cookbook(self, cookbook: str, args: Sequence[str] = ()) -> int:
        """Run another Cookbook within the current run.

        The other Cookbook will be executed with the current setup and will log in the same file of the current
        Cookbook that is running.

        Arguments:
            cookbook: the path to the cookbook to execute, either in Spicerack's dot notation or the relative path to
                the Python file to execute.
            args: an iterable sequence of strings with the Cookbook's argument. The Cookbook will be executed with the
                same global arguments used for the current run.

        Returns:
            The exit code of the Cookbook, 0 if successful, non-zero if not.

        Raises:
            spicerack.exceptions.SpicerackError: if the ``get_cookbook_callback`` callback is not set or unable to find
                the cookbook with the given name.

        """
        if self._get_cookbook_callback is None:
            raise SpicerackError("Unable to run other cookbooks, get_cookbook_callback is not set.")

        cookbook_item = self._get_cookbook_callback(self, cookbook, args)
        if cookbook_item is None:
            raise SpicerackError(f"Unable to find cookbook {cookbook}")

        logger.debug("Executing cookbook %s with args: %s", cookbook, args)
        return cookbook_item.run() or 0  # Force the return code to be 0 if the cookbook returns None

    def remote(self, installer: bool = False) -> Remote:
        """Get a Remote instance.

        Arguments:
            installer: whether to use the special configuration to connect to a Debian installer or freshly re-imaged
                host prior to its first Puppet run.

        """
        return Remote(self._cumin_installer_config if installer else self._cumin_config, dry_run=self._dry_run)

    def confctl(self, entity_name: str) -> ConftoolEntity:
        """Get a Conftool specific entity instance.

        Arguments:
            entity_name: the name of a Conftool entity. Possible values: ``node``, ``service``, ``discovery``,
                ``mwconfig``.

        """
        if self._confctl is None:
            self._confctl = Confctl(
                config=self._conftool_config,
                schema=self._conftool_schema,
                dry_run=self._dry_run,
            )

        return self._confctl.entity(entity_name)

    def dhcp(self, remote_hosts: RemoteHosts) -> DHCP:
        """Return a DHCP configuration manager for the specified site.

        Arguments:
            remote_hosts: Hosts to operate on, which are normally install servers with dhcp.

        """
        return DHCP(remote_hosts, dry_run=self._dry_run)

    def dns(self) -> Dns:
        """Get a Dns instance that will use the operating system default nameserver(s)."""
        return Dns()

    def discovery(self, *records: str) -> Discovery:
        """Get a Discovery instance for the given records.

        Arguments:
            *records: arbitrary positional arguments, each one must be a Discovery DNS record name.

        """
        return Discovery(
            conftool=self.confctl("discovery"),
            authdns_servers=self.authdns_servers,
            records=list(records),
            dry_run=self._dry_run,
        )

    def kubernetes(self, group: str, cluster: str) -> Kubernetes:
        """Get a kubernetes client for the specified cluster.

        Arguments:
            group: the cluster group.
            cluster: the kubernetes cluster.

        """
        return Kubernetes(group, cluster, dry_run=self._dry_run)

    def mediawiki(self) -> MediaWiki:
        """Get a MediaWiki instance."""
        return MediaWiki(
            self.confctl("mwconfig"),
            self.remote(),
            self._username,
            dry_run=self._dry_run,
        )

    def mysql(self) -> Mysql:
        """Get a Mysql instance."""
        return Mysql(dry_run=self._dry_run)

    def mysql_legacy(self) -> MysqlLegacy:
        """Get a MysqlLegacy instance."""
        return MysqlLegacy(self.remote(), dry_run=self._dry_run)

    def redis_cluster(self, cluster: str) -> RedisCluster:
        """Get a RedisCluster instance.

        Arguments:
            cluster: the name of the cluster.

        """
        return RedisCluster(
            cluster,
            self._spicerack_config_dir / "redis_cluster",
            dry_run=self._dry_run,
        )

    def reposync(self, name: str) -> RepoSync:
        """Get a Reposync instance.

        Arguments:
            name: the name of the repo to sync.

        """
        config = load_yaml_config(self._spicerack_config_dir / "reposync" / "config.yaml")
        if name not in config["repos"]:
            raise SpicerackError(f"Unknown repo {name}")

        repo_dir = Path(config["base_dir"], name)
        query = ",".join(config["remotes"])
        remote_hosts = self.remote().query(query)

        if not repo_dir.is_dir():
            raise SpicerackError(f"The repo directory ({repo_dir}) does not exist")
        repo = Repo(repo_dir)
        if not repo.bare:
            raise SpicerackError(f"The repo directory ({repo_dir}) is not a bare git repository")

        return RepoSync(repo, self.username, remote_hosts, dry_run=self._dry_run)

    def elasticsearch_clusters(
        self, clustergroup: str, write_queue_datacenters: Sequence[str]
    ) -> ElasticsearchClusters:
        """Get an ElasticsearchClusters instance.

        Arguments:
            clustergroup: name of cluster group e.g ``search_eqiad``.
            write_queue_datacenters: Sequence of which core DCs to query write queues for.

        """
        configuration = load_yaml_config(self._spicerack_config_dir / "elasticsearch" / "config.yaml")

        return create_elasticsearch_clusters(
            configuration,
            clustergroup,
            write_queue_datacenters,
            self.remote(),
            self.prometheus(),
            dry_run=self._dry_run,
        )

    def admin_reason(self, reason: str, task_id: Optional[str] = None) -> Reason:
        """Get an administrative Reason instance.

        Arguments:
            reason: the reason to use to justify an administrative action. See
                :py:class:`spicerack.administrative.Reason` for all the details.
            task_id: the task ID to mention in the reason.

        """
        return Reason(reason, self._username, self._current_hostname, task_id=task_id)

    def icinga_hosts(self, target_hosts: TypeHosts, *, verbatim_hosts: bool = False) -> IcingaHosts:
        """Get an IcingaHosts instance.

        Note:
            To interact with both Icinga and Alertmanager alerts, use
            :py:meth:`spicerack.Spicerack.alerting_hosts` instead.

        Arguments:
            target_hosts: the target hosts either as a NodeSet instance or a sequence of strings.
            verbatim_hosts: if :py:data:`True` use the hosts passed verbatim as is, if instead :py:data:`False`, the
                default, consider the given target hosts as FQDNs and extract their hostnames to be used in Icinga.

        """
        return IcingaHosts(self.icinga_master_host, target_hosts, verbatim_hosts=verbatim_hosts, dry_run=self._dry_run)

    def puppet(self, remote_hosts: RemoteHosts) -> PuppetHosts:
        """Get a PuppetHosts instance for the given remote hosts.

        Arguments:
            remote_hosts: the instance with the target hosts.

        """
        return PuppetHosts(remote_hosts)

    def puppet_server(self) -> PuppetServer:
        """Get a PuppetServer instance to manage hosts and certificates from a Puppet master."""
        domain = "_x-puppet-ca._tcp.eqiad.wmnet"
        response = self.dns().resolve(domain, "SRV")
        if response is None or response.rrset is None:
            raise SpicerackError(f"Unable to find record for {domain}")
        answers = [str(rdata.target) for rdata in response.rrset]
        if not answers:
            raise SpicerackError("Unable to find any ca servers from DNS")
        if len(answers) > 1:
            raise SpicerackError(f"Found multiple ca servers from DNS: {','.join(answers)}")
        return PuppetServer(self.remote().query(answers[0].rstrip(".")))

    def puppet_master(self) -> PuppetMaster:
        """Get a PuppetMaster instance to manage hosts and certificates from a Puppet master."""
        return PuppetMaster(self.remote().query(get_puppet_ca_hostname()))

    def ipmi(self, target: str) -> Ipmi:
        """Get an Ipmi instance to send remote IPMI commands to management consoles.

        Arguments:
            target: the management console FQDN or IP to target.

        """
        return Ipmi(target, self.management_password, dry_run=self._dry_run)

    def phabricator(self, bot_config_file: str, section: str = "phabricator_bot") -> Phabricator:
        """Get a Phabricator instance to interact with a Phabricator website.

        Arguments:
            bot_config_file: the path to the configuration file for the Phabricator bot, with the following structure::

                    [section_name]
                    host = https://phabricator.example.com/api/
                    username = phab-bot
                    token = api-12345

            section: the name of the section of the configuration file where to find the required parameters.

        """
        # Allow to specify the configuration file as opposed to other methods so that different clients can use
        # different Phabricator BOT accounts, potentially with different permissions.
        return create_phabricator(bot_config_file, section=section, dry_run=self._dry_run)

    def prometheus(self) -> Prometheus:
        """Get a Prometheus instance."""
        return Prometheus()

    def thanos(self) -> Thanos:
        """Get a Thanos instance."""
        return Thanos()

    def debmonitor(self) -> Debmonitor:
        """Get a Debmonitor instance to interact with a Debmonitor website.

        Raises:
            KeyError: if any configuration option is missing.

        """
        options = load_ini_config(self._debmonitor_config).defaults()
        return Debmonitor(options["server"], options["cert"], options["key"], dry_run=self._dry_run)

    def ganeti(self) -> Ganeti:
        """Get an instance to interact with Ganeti.

        Raises:
            KeyError: If the configuration file does not contain the correct keys.

        """
        configuration = load_yaml_config(self._spicerack_config_dir / "ganeti" / "config.yaml")

        return Ganeti(
            configuration["username"],
            configuration["password"],
            configuration["timeout"],
            self.remote(),
            self.netbox(),
        )

    def netbox(self, *, read_write: bool = False) -> Netbox:
        """Get a Netbox instance to interact with Netbox's API.

        Arguments:
            read_write: whether to use a read-write token.

        """
        config = load_yaml_config(self._spicerack_config_dir / "netbox" / "config.yaml")
        if read_write and not self._dry_run:
            token = config["api_token_rw"]
        else:
            token = config["api_token_ro"]

        return Netbox(config["api_url"], token, dry_run=self._dry_run)

    def netbox_server(self, hostname: str, *, read_write: bool = False) -> NetboxServer:
        """Get a NetboxServer instance to interact with a server in Netbox, both physical and virtual.

        Arguments:
            hostname: the hostname (not FQDN) of the server to manage.
            read_write: whether to use a read-write token.

        Raises:
            spicerack.netbox.NetboxError: if unable to find or load the server.

        """
        return self.netbox(read_write=read_write).get_server(hostname)

    def requests_session(self, name: str, **kwargs: Any) -> requests.Session:
        """Return a new requests Session with timeout and retry logic.

        Params:
            according to :py:func:`wmflib.requests.http_session`.

        """
        name = f"Spicerack/{__version__} {name}"
        return requests.http_session(name, **kwargs)

    def etcdctl(self, *, remote_host: RemoteHosts) -> EtcdctlController:
        """Add etcdctl control capabilities to the given RemoteHost.

        Params:
            remote_host: Single remote host (that should map one single host)
                to add capabilities to.

        Returns:
            A wrapped RemoteHost with the etcdctl control related methods.

        """
        return EtcdctlController(remote_host=remote_host)

    def kafka(self) -> Kafka:
        """Get an instance to interact with Kafka.

        Raises:
            KeyError: If the configuration file does not contain the correct keys.

        """
        configuration = load_yaml_config(self._spicerack_config_dir / "kafka" / "config.yaml")

        return Kafka(kafka_config=configuration, dry_run=self._dry_run)

    def redfish(self, hostname: str, username: str = "root", password: str = "") -> Redfish:  # nosec
        """Get an instance to talk to the Redfish API of a physical server.

        Notes:
            At the moment only Dell hardware is supported.

        Arguments:
            hostname: the hostname (not FQDN) of the physical server to manage.
            username: the username for the management console.
            password: the password for the management console for the given user. If empty or not provided would use
                the production management password and ask the user for it if not already in memory.

        Raises:
            spicerack.exceptions.SpicerackError: if not a physical server or unable to find the management IP.

        """
        if not password:
            password = self.management_password

        netbox = self.netbox()
        if netbox.get_server(hostname).virtual:
            raise SpicerackError(f"Host {hostname} is not a Physical server, Redfish is not supported.")

        netbox_ip = netbox.api.ipam.ip_addresses.get(device=hostname, interface=MANAGEMENT_IFACE_NAME)

        # TODO: generalize when support for additional vendors will be added.
        return RedfishDell(hostname, ip_interface(netbox_ip), username, password, dry_run=self._dry_run)

    def alertmanager_hosts(self, target_hosts: TypeHosts, *, verbatim_hosts: bool = False) -> AlertmanagerHosts:
        """Get an AlertmanagerHosts instance.

        Note:
            To interact with both Icinga and Alertmanager alerts, use
            :py:meth:`spicerack.Spicerack.alerting_hosts` instead.

        Arguments:
            target_hosts: the target hosts either as a NodeSet instance or a sequence of strings.
            verbatim_hosts: if :py:data:`True` use the hosts passed verbatim as is, if instead :py:data:`False`, the
                default, consider the given target hosts as FQDNs and extract their hostnames to be used in Icinga.

        """
        return AlertmanagerHosts(target_hosts, verbatim_hosts=verbatim_hosts, dry_run=self._dry_run)

    def alertmanager(self) -> Alertmanager:
        """Get an Alertmanager instance.

        Note:
            To interact with Alertmanager alerts attached to an ``instance`` use
            :py:meth:`spicerack.Spicerack.alertmanager_hosts` instead.

        """
        return Alertmanager(dry_run=self._dry_run)

    def alerting_hosts(self, target_hosts: TypeHosts, *, verbatim_hosts: bool = False) -> AlertingHosts:
        """Get an AlertingHosts instance.

        Arguments:
            target_hosts: the target hosts either as a NodeSet instance or a sequence of strings.
            verbatim_hosts: if :py:data:`True` use the hosts passed verbatim as is, if instead :py:data:`False`, the
                default, consider the given target hosts as FQDNs and extract their hostnames to be used in Icinga.

        """
        return AlertingHosts(
            self.alertmanager_hosts(target_hosts, verbatim_hosts=verbatim_hosts),
            self.icinga_hosts(target_hosts, verbatim_hosts=verbatim_hosts),
        )

    def service_catalog(self) -> Catalog:
        """Get a Catalog instance that reflects Puppet's service::catalog hieradata variable."""
        if self._service_catalog is None:
            config = load_yaml_config(self._spicerack_config_dir / "service" / "service.yaml")
            self._service_catalog = Catalog(
                config, confctl=self.confctl("discovery"), authdns_servers=self.authdns_servers, dry_run=self._dry_run
            )

        return self._service_catalog

    def peeringdb(self, *, ttl: int = 86400) -> PeeringDB:
        """Get a PeeringDB instance to interact with the PeeringDB API.

        Arguments:
            ttl: the cache timeout in seconds. If cached items are older than the given TTL they will be ignored and
                fetched again.

        """
        config = load_yaml_config(self._spicerack_config_dir / "peeringdb" / "config.yaml")
        token = config.get("api_token_ro", "")
        cachedir = config.get("cachedir")
        if cachedir is not None:
            cachedir = Path(cachedir)

        return PeeringDB(cachedir=cachedir, ttl=ttl, proxies=self.requests_proxies, token=token)

    def apt_get(self, remote_hosts: RemoteHosts) -> AptGetHosts:
        """Get an APTGet instance for the given remote hosts.

        Examples:
            ::

                >>> hosts = spicerack.remote().query('A:myalias')
                >>> apt_get = spicerack.apt_get(hosts)

        Arguments:
            remote_hosts: the instance with the target hosts.

        """
        return AptGetHosts(remote_hosts)


class SpicerackExtenderBase:
    """Base class to create a Spicerack extender. Necessary when the ``external_modules_dir`` configuration is set."""

    def __init__(self, *, spicerack: Spicerack):
        """Initialize the instance.

        Arguments:
            spicerack: the Spicerack instance.

        """
        self._spicerack = spicerack



"""Class API example cookbook."""
import argparse

from spicerack.cookbook import CookbookBase, CookbookRunnerBase


class ExampleCookbook(CookbookBase):  # noqa: D101
    # No docstring to test the spicerack.cookbook.CookbookBase.title implementation

    def argument_parser(self):
        """As defined by the parent class."""
        return argparse.ArgumentParser("Argparse")

    def get_runner(self, args):
        """As required by the parent class."""
        return ExampleRunner()


class ExampleRunner(CookbookRunnerBase):
    """As required by spicerack._cookbook."""

    def run(self):
        """As required by the parent class."""



# -*- coding: utf-8 -*-
"""Sphynx configuration."""
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use pathlib's resolve() to make it absolute, like shown here.
#
import os
import sys
from datetime import date
from pathlib import Path

import sphinx_rtd_theme
from pkg_resources import get_distribution

# Adjust path
sys.path.insert(0, Path(__file__).parent.parent.resolve())

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
    "sphinx_autodoc_typehints",
    "sphinxarg.ext",
]
if not os.environ.get("PYBUILD_NAME", ""):
    extensions.append("sphinxcontrib.jquery")

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "Spicerack"
title = f"{project} Documentation"
copyright = f"2018-{date.today().year}, Riccardo Coccioli <rcoccioli@wikimedia.org>, Wikimedia Foundation, Inc."
author = "Riccardo Coccioli"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = get_distribution("wikimedia-spicerack").version
# The short X.Y version.
version = release


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "Spicerackdoc"


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ("cookbook", "cookbook", "Spicerack Cookbook Runner", [author], 1),
]


# -- Options for intersphinx ---------------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "ClusterShell": ("https://clustershell.readthedocs.io/en/v1.8.1/", None),
    "cumin": ("https://doc.wikimedia.org/cumin/master/", None),
    "requests": ("https://requests.readthedocs.io/en/stable/", None),
    "pymysql": ("https://pymysql.readthedocs.io/en/latest/", None),
    "wmflib": ("https://doc.wikimedia.org/wmflib/master/", None),
    "kafka": ("https://kafka-python.readthedocs.io/en/stable/", None),
    "git": ("https://gitpython.readthedocs.io/en/stable/", None),
    "kubernetes": ("https://kubernetes.readthedocs.io/en/latest/", None),
}

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_keyword = True
napoleon_use_rtype = True
napoleon_type_aliases = None
napoleon_attr_annotations = True

# Type hints settings
typehints_fully_qualified = True
always_document_param_types = False
typehints_document_rtype = True
typehints_use_rtype = True
typehints_defaults = "comma"
typehints_use_signature = True
typehints_use_signature_return = True

# Autodoc settings
autodoc_default_options = {
    # Using None as value instead of True to support the version of Sphinx used in Buster
    "members": None,
    "member-order": "groupwise",
    "show-inheritance": None,
}
autoclass_content = "both"


# -- Helper functions -----------------------------------------------------


def filter_namedtuple_docstrings(app, what, name, obj, options, lines):
    """Fix the automatically generated docstrings for namedtuples classes."""
    if what == "property" and len(lines) == 1 and lines[0].startswith("Alias for field number"):
        del lines[:]


# Keep track of documented classes to avoid annotating both class and __init__.
# Necessary when using autoclass_content 'both' and add_abstract_annotations().
_spicerack_documented_classes = set()


def add_abstract_annotations(app, what, name, obj, options, lines):
    """Workaround to add an abstract annotation for ABC abstract classes."""
    if (
        what == "class"
        and len(getattr(obj, "__abstractmethods__", [])) > 0
        and name not in _spicerack_documented_classes
    ):
        lines.insert(0, "``abstract``")
        _spicerack_documented_classes.add(name)


def setup(app):
    """Register the filter_namedtuple_docstrings function."""
    app.connect("autodoc-process-docstring", filter_namedtuple_docstrings)
    app.connect("autodoc-process-docstring", add_abstract_annotations)
    app.add_css_file("theme_overrides.css")  # override wide tables in RTD theme



"""Alertmanager module tests."""
import logging
from datetime import datetime, timedelta, timezone
from unittest import mock

import pytest
import requests
from cumin import nodeset

from spicerack import alertmanager
from spicerack.administrative import Reason


class TestAlertmanager:
    """Tests for the Alertmanager class."""

    @pytest.fixture(autouse=True)
    def setup_method(self, requests_mock):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        self.matchers = [
            {"name": "site", "value": "dc1", "isRegex": False},
            {"name": "label1", "value": "value1", "isRegex": False},
        ]
        self.alertmanager = alertmanager.Alertmanager(dry_run=False)
        self.am_dry_run = alertmanager.Alertmanager(dry_run=True)
        self.requests_mock = requests_mock
        self.reason = Reason("test", "user", "host")

    def test_add_silence_basic(self):
        """It should issue a silence with all the matchers."""
        self.requests_mock.post("/api/v2/silences", json={"silenceID": "foobar"})
        response = self.alertmanager.downtime(self.reason, matchers=self.matchers)
        assert response == "foobar"
        assert self.requests_mock.last_request.hostname == "alertmanager-eqiad.wikimedia.org"
        request_json = self.requests_mock.last_request.json()
        assert request_json["matchers"] == self.matchers
        assert request_json["comment"] == "test - user@host"
        assert request_json["createdBy"] == "user@host"

    def test_add_silence_no_matchers(self):
        """It should raise an AlertmanagerError if there are no matchers specified."""
        with pytest.raises(alertmanager.AlertmanagerError, match="No matchers provided"):
            self.alertmanager.downtime(self.reason, matchers=[])

    def test_add_silence_duration(self):
        """It should issue a silence with a given duration."""
        self.requests_mock.post("/api/v2/silences", json={"silenceID": "foobar"})
        with mock.patch("spicerack.alertmanager.datetime") as dt_mock:
            dt_mock.utcnow.return_value = datetime(2022, 6, 6, 10, 00, 00, tzinfo=timezone.utc)
            self.alertmanager.downtime(self.reason, matchers=self.matchers, duration=timedelta(hours=6))
        request_json = self.requests_mock.last_request.json()
        assert request_json["startsAt"] == "2022-06-06T10:00:00+00:00"
        assert request_json["endsAt"] == "2022-06-06T16:00:00+00:00"

    def test_add_silence_dry_run(self):
        """It should not create a silence because in dry-run mode."""
        self.requests_mock.post("/api/v2/silences", json={"silenceID": "foobar"})
        response = self.am_dry_run.downtime(self.reason, matchers=self.matchers)
        assert response == ""
        assert self.requests_mock.call_count == 0

    def test_delete_silence_dry_run(self):
        """It should delete a downtime."""
        self.requests_mock.delete("/api/v2/silence/foobar")
        self.am_dry_run.remove_downtime("nonexistent")
        assert self.requests_mock.call_count == 0

    def test_delete_silence_basic(self):
        """It should delete a downtime."""
        self.requests_mock.delete("/api/v2/silence/foobar")
        self.alertmanager.remove_downtime("foobar")
        assert self.requests_mock.call_count == 1

    def test_delete_silence_already_deleted(self, caplog):
        """It should not error if the downtime has been already deleted or is expired."""
        self.requests_mock.delete("/api/v2/silence/foobar", status_code=500, json="silence foobar already expired")
        with caplog.at_level(logging.WARNING):
            self.alertmanager.remove_downtime("foobar")

        assert "Silence ID foobar has been already deleted or is expired" in caplog.text
        assert self.requests_mock.call_count == 2

    def test_delete_silence_error(self, caplog):
        """It should raise an AlertmanagerError on any other error."""
        self.requests_mock.delete("/api/v2/silence/foobar", status_code=500, json="silence not found")
        with caplog.at_level(logging.WARNING):
            with pytest.raises(alertmanager.AlertmanagerError, match="Unable to DELETE to any Alertmanager"):
                self.alertmanager.remove_downtime("foobar")

        assert "already deleted" not in caplog.text
        assert self.requests_mock.call_count == 2

    def test_downtimed(self):
        """It should issue a silence and then delete it."""
        self.requests_mock.post("/api/v2/silences", json={"silenceID": "foobar"})
        self.requests_mock.delete("/api/v2/silence/foobar")
        with self.alertmanager.downtimed(self.reason, matchers=self.matchers):
            assert self.requests_mock.call_count == 1
        assert self.requests_mock.call_count == 2

    @pytest.mark.parametrize(
        "remove_on_error, total_call_count",
        (
            (True, 2),
            (False, 1),
        ),
    )
    def test_downtimed_remove_on_error(self, remove_on_error, total_call_count):
        """It should issue a silence and then delete it even with errors."""
        self.requests_mock.post("/api/v2/silences", json={"silenceID": "foobar"})
        self.requests_mock.delete("/api/v2/silence/foobar")
        with pytest.raises(ValueError):
            with self.alertmanager.downtimed(self.reason, matchers=self.matchers, remove_on_error=remove_on_error):
                assert self.requests_mock.call_count == 1
                raise ValueError()
        assert self.requests_mock.call_count == total_call_count

    def test_connection_errors(self):
        """It should raise AlertmanagerError when unable to contact the API."""
        self.requests_mock.post("/api/v2/silences", exc=requests.exceptions.ConnectionError)
        with pytest.raises(alertmanager.AlertmanagerError):
            self.alertmanager.downtime(self.reason, matchers=self.matchers)

    def test_fallback_on_error(self):
        """It should fallback to the next Alertmanager on error."""
        ams = alertmanager.ALERTMANAGER_URLS
        self.requests_mock.post(f"{ams[0]}/api/v2/silences", exc=requests.exceptions.ConnectionError)
        self.requests_mock.post(f"{ams[1]}/api/v2/silences", json={"silenceID": "foobar"})
        assert "foobar" == self.alertmanager.downtime(self.reason, matchers=self.matchers)
        assert self.requests_mock.last_request.hostname == "alertmanager-codfw.wikimedia.org"


class TestAlertmanagerHosts:
    """Tests for the AlertmanagerHosts class."""

    @pytest.fixture(autouse=True)
    def setup_method(self, requests_mock):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        self.am_hosts = alertmanager.AlertmanagerHosts(["host1", "host2"], dry_run=False)
        self.am_hosts_dry_run = alertmanager.AlertmanagerHosts(["host1", "host2"], dry_run=True)
        self.requests_mock = requests_mock
        self.reason = Reason("test", "user", "host")

    @pytest.mark.parametrize(
        "hosts, regex",
        (
            (["host1", "host2"], r"^(host1|host2)(\..+)?(:[0-9]+)?$"),
            (["host1:1234", "host2"], r"^(host1:1234|host2(\..+)?(:[0-9]+)?)$"),
            (["host1.example.com:1234", "host2"], r"^(host1|host2)(\..+)?(:[0-9]+)?$"),
            (["host1:1234", "host2.example.com:5678"], r"^(host1:1234|host2(\..+)?(:[0-9]+)?)$"),
        ),
    )
    def test_add_silence_basic(self, hosts, regex):
        """It should issue a silence with all defaults."""
        self.requests_mock.post("/api/v2/silences", json={"silenceID": "foobar"})
        am_hosts = alertmanager.AlertmanagerHosts(hosts, dry_run=False)
        response = am_hosts.downtime(self.reason)
        assert response == "foobar"
        assert self.requests_mock.last_request.hostname == "alertmanager-eqiad.wikimedia.org"
        request_json = self.requests_mock.last_request.json()
        assert request_json["matchers"] == [
            {"name": "instance", "value": regex, "isRegex": True},
        ]
        assert request_json["comment"] == "test - user@host"
        assert request_json["createdBy"] == "user@host"

    def test_add_silence_additional_matchers(self):
        """It should issue a silence with the provided additional matchers."""
        self.requests_mock.post("/api/v2/silences", json={"silenceID": "foobar"})
        self.am_hosts.downtime(self.reason, matchers=({"name": "severity", "value": "critical", "isRegex": False},))
        request_json = self.requests_mock.last_request.json()
        assert request_json["matchers"] == [
            {"name": "severity", "value": "critical", "isRegex": False},
            {"name": "instance", "value": r"^(host1|host2)(\..+)?(:[0-9]+)?$", "isRegex": True},
        ]

    def test_add_silence_additional_matchers_invalid(self):
        """It should raise an AlertmanagerError if any of the matchers target the instance property."""
        with pytest.raises(alertmanager.AlertmanagerError, match="Matchers cannot target the instance property"):
            self.am_hosts.downtime(self.reason, matchers=({"name": "instance", "value": "host1001", "isRegex": False},))

    def test_add_silence_port_included(self):
        """It should issue a silence with the specific port and not any port in the matcher."""
        self.requests_mock.post("/api/v2/silences", json={"silenceID": "foobar"})
        am_hosts = alertmanager.AlertmanagerHosts(["host1:1234", "host2:5678"], dry_run=False)
        am_hosts.downtime(self.reason)
        request_json = self.requests_mock.last_request.json()
        assert request_json["matchers"] == [
            {"name": "instance", "value": r"^(host1:1234|host2:5678)$", "isRegex": True},
        ]

    def test_add_silence_duration(self):
        """It should issue a silence with a given duration."""
        self.requests_mock.post("/api/v2/silences", json={"silenceID": "foobar"})
        with mock.patch("spicerack.alertmanager.datetime") as dt_mock:
            dt_mock.utcnow.return_value = datetime(2022, 6, 6, 10, 00, 00, tzinfo=timezone.utc)
            self.am_hosts.downtime(self.reason, duration=timedelta(hours=6))
        request_json = self.requests_mock.last_request.json()
        assert request_json["startsAt"] == "2022-06-06T10:00:00+00:00"
        assert request_json["endsAt"] == "2022-06-06T16:00:00+00:00"

    def test_add_silence_dry_run(self):
        """It should not create a silence because in dry-run mode."""
        self.requests_mock.post("/api/v2/silences", json={"silenceID": "foobar"})
        response = self.am_hosts_dry_run.downtime(self.reason)
        assert response == ""
        assert self.requests_mock.call_count == 0

    def test_verbatim_hosts(self):
        """It should issue silences for verbatim hosts."""
        self.requests_mock.post("/api/v2/silences", json={"silenceID": "foobar"})
        am_hosts = alertmanager.AlertmanagerHosts(
            ["host1.foo.bar", "host2.bar.baz:1234"], verbatim_hosts=True, dry_run=False
        )
        am_hosts.downtime(self.reason)
        request_json = self.requests_mock.last_request.json()
        assert request_json["matchers"] == [
            {
                "name": "instance",
                "value": r"^(host1\.foo\.bar(\..+)?(:[0-9]+)?|host2\.bar\.baz:1234)$",
                "isRegex": True,
            },
        ]

    def test_nodeset_hosts(self):
        """It should expand NodeSet hosts."""
        self.requests_mock.post("/api/v2/silences", json={"silenceID": "foobar"})
        am_hosts = alertmanager.AlertmanagerHosts(nodeset("host[1-2]"), verbatim_hosts=True, dry_run=False)
        am_hosts.downtime(self.reason)
        request_json = self.requests_mock.last_request.json()
        assert request_json["matchers"] == [
            {"name": "instance", "value": r"^(host1|host2)(\..+)?(:[0-9]+)?$", "isRegex": True},
        ]

    def test_empty_target_hosts(self):
        """It should error with empty hosts."""
        with pytest.raises(alertmanager.AlertmanagerError):
            alertmanager.AlertmanagerHosts([""])

    def test_downtimed(self):
        """It should issue a silence and then delete it."""
        self.requests_mock.post("/api/v2/silences", json={"silenceID": "foobar"})
        self.requests_mock.delete("/api/v2/silence/foobar")
        with self.am_hosts.downtimed(self.reason):
            assert self.requests_mock.call_count == 1
        assert self.requests_mock.call_count == 2

    @pytest.mark.parametrize(
        "remove_on_error, total_call_count",
        (
            (True, 2),
            (False, 1),
        ),
    )
    def test_downtimed_remove_on_error(self, remove_on_error, total_call_count):
        """It should issue a silence and then delete it even with errors."""
        self.requests_mock.post("/api/v2/silences", json={"silenceID": "foobar"})
        self.requests_mock.delete("/api/v2/silence/foobar")
        with pytest.raises(ValueError):
            with self.am_hosts.downtimed(self.reason, remove_on_error=remove_on_error):
                assert self.requests_mock.call_count == 1
                raise ValueError()
        assert self.requests_mock.call_count == total_call_count



"""Netbox module tests."""

from unittest import mock

import pynetbox
import pytest

from spicerack.netbox import Netbox, NetboxAPIError, NetboxError, NetboxHostNotFoundError, NetboxServer

NETBOX_URL = "https://example.com/"
NETBOX_TOKEN = "secret_token"


def _request_error():
    """Return a Netbox RequestError."""
    fakestatus = mock.Mock()
    fakestatus.status_code = 404
    return pynetbox.RequestError(fakestatus)


def _base_netbox_host(name):
    host = mock.MagicMock()
    del host.keys  # Allow to call dict() on this object
    host.name = name
    host.__str__.return_value = name
    host.asset_tag = "ASSET1234"
    host.status.__str__.return_value = "Active"
    host.status.value = "active"
    host.serialize.return_value = {"name": name}
    host.save.return_value = True
    host.primary_ip4.dns_name = f"{name}.example.com"
    host.primary_ip6.dns_name = f"{name}.example.com"

    dict_repr = {
        "name": name,
        "asset_tag": "ASSET1234",
        "status": {"value": host.status.value, "label": str(host.status)},
        "role": {"id": 1, "name": host.role.name, "slug": host.role.slug},
        "primary_ip4": {
            "id": 1,
            "family": 4,
            "address": "10.0.0.1/22",
            "dns_name": host.primary_ip4.dns_name,
        },
        "primary_ip6": {
            "id": 1,
            "family": 6,
            "address": "2620:0:861:103:10:0:0:1/64",
            "dns_name": host.primary_ip6.dns_name,
        },
        "cluster": host.cluster,
    }

    return host, dict_repr


@pytest.fixture(name="netbox_host")
def _netbox_host():
    """Return a mocked Netbox physical device."""
    host, dict_repr = _base_netbox_host("physical")
    host.cluster = None  # A physical server does not belong to a VM cluster
    host.device_role.slug = "server"
    host.device_role.name = "Server"
    dict_repr["cluster"] = None
    host.__iter__.return_value = dict_repr.items()
    return host


@pytest.fixture(name="netbox_virtual_machine")
def _netbox_virtual_machine():
    """Return a mocked Netbox virtual machine."""
    host, dict_repr = _base_netbox_host("virtual")
    host.cluster.name = "testcluster"
    host.role.slug = "server"
    host.role.name = "Server"
    dict_repr["cluster"] = {"id": 1, "name": host.cluster.name}
    del host.rack  # A virtual machine doesn't have a rack property
    host.__iter__.return_value = dict_repr.items()

    return host


class TestNetbox:
    """Tests for the Netbox class."""

    @mock.patch("pynetbox.api")
    def setup_method(self, _, mocked_api):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_api = mocked_api
        self.netbox = Netbox(NETBOX_URL, NETBOX_TOKEN, dry_run=False)
        self.netbox_dry_run = Netbox(NETBOX_URL, NETBOX_TOKEN, dry_run=True)

    def test_netbox_api(self):
        """An instance of Netbox should instantiate the Netbox API and expose it via the api property."""
        self.mocked_api.called_once_with(NETBOX_URL, token=NETBOX_TOKEN)
        assert self.netbox.api == self.mocked_api()

    def test_netbox_fetch_host_status_nohost(self):
        """Test the error scenario where the host is not found."""
        self.mocked_api().dcim.devices.get.return_value = None
        self.mocked_api().virtualization.virtual_machines.get.return_value = None
        with pytest.raises(NetboxHostNotFoundError):
            self.netbox.fetch_host_status("host")

    def test_netbox_fetch_host_status_error(self):
        """Test the error scenario where the host is not found."""
        self.mocked_api().dcim.devices.get = mock.Mock(side_effect=_request_error())
        self.mocked_api().virtualization.virtual_machines.get = mock.Mock(side_effect=_request_error())
        with pytest.raises(NetboxError, match="Error retrieving Netbox host"):
            self.netbox.fetch_host_status("host")

    def test_fetch_host_status(self, netbox_host):
        """Test fetching host status."""
        self.mocked_api().dcim.devices.get.return_value = netbox_host
        self.mocked_api().virtualization.virtual_machines.get.return_value = None

        assert self.netbox.fetch_host_status("host") == "Active"

    def test_put_host_status_good_status(self, netbox_host):
        """Test setting a status and it working."""
        self.mocked_api().dcim.devices.get.return_value = netbox_host
        self.mocked_api().virtualization.virtual_machines.get.return_value = None

        self.netbox.put_host_status("host", "Planned")
        assert str(netbox_host.status) == "planned"
        assert netbox_host.save.called

    def test_put_host_status_save_failure(self, netbox_host):
        """Test save failure."""
        self.mocked_api().dcim.devices.get.return_value = netbox_host
        self.mocked_api().virtualization.virtual_machines.get.return_value = None
        netbox_host.save.return_value = False
        with pytest.raises(
            NetboxAPIError,
            match="Failed to update Netbox status for host physical Active -> planned",
        ):
            self.netbox.put_host_status("physical", "Planned")

        assert netbox_host.save.called

    def test_put_host_status_dry_run_success(self, netbox_host):
        """Test dry run, which  should always work on save."""
        self.mocked_api().dcim.devices.get.return_value = netbox_host
        netbox_host.save.return_value = False
        self.netbox_dry_run.put_host_status("physical", "Planned")
        netbox_host.save.assert_not_called()

    def test_put_host_status_error(self, netbox_host):
        """Test exception during save."""
        self.mocked_api().dcim.devices.get.return_value = netbox_host
        self.mocked_api().virtualization.virtual_machines.get.return_value = None
        netbox_host.save.side_effect = _request_error()
        with pytest.raises(
            NetboxAPIError,
            match="Failed to save Netbox status for host physical Active -> planned",
        ):
            self.netbox.put_host_status("physical", "Planned")

        assert netbox_host.save.called

    def test_fetch_host_detail(self, netbox_host):
        """Test fetching host detail."""
        self.mocked_api().dcim.devices.get.return_value = netbox_host
        self.mocked_api().virtualization.virtual_machines.get.return_value = None
        detail = self.netbox.fetch_host_detail("physical")
        assert netbox_host.serialize.called
        assert not detail["is_virtual"]

    def test_fetch_host_detail_vm(self, netbox_virtual_machine):
        """Virtual machines should have is_virtual == True and a cluster name."""
        self.mocked_api().dcim.devices.get.return_value = None
        self.mocked_api().virtualization.virtual_machines.get.return_value = netbox_virtual_machine
        detail = self.netbox.fetch_host_detail("virtual")
        assert netbox_virtual_machine.serialize.called
        assert detail["is_virtual"]
        assert detail["ganeti_cluster"] == "testcluster"

    def test_fetch_host_status_vm(self, netbox_virtual_machine):
        """Virtual machines should return status just like devices."""
        self.mocked_api().dcim.devices.get.return_value = None
        self.mocked_api().virtualization.virtual_machines.get.return_value = netbox_virtual_machine
        assert self.netbox.fetch_host_status("virtual") == "Active"

    def test_fetch_host_status_vm_error(self):
        """Virtual machines should raise an exception on an API error."""
        self.mocked_api().dcim.devices.get.return_value = None
        self.mocked_api().virtualization.virtual_machines.get = mock.Mock(side_effect=_request_error())
        with pytest.raises(NetboxAPIError, match="Error retrieving Netbox VM"):
            self.netbox.fetch_host_status("virtual")

    def test_set_host_status_vm(self, netbox_virtual_machine):
        """Virtual machines should raise an exception if you try to set the status."""
        self.mocked_api().dcim.devices.get.return_value = None
        self.mocked_api().virtualization.virtual_machines.get.return_value = netbox_virtual_machine

        with pytest.raises(NetboxHostNotFoundError):
            self.netbox.put_host_status("virtual", "Active")

        netbox_virtual_machine.save.assert_not_called()

    def test_get_server_physical(self, netbox_host):
        """It should return the NetboxServer instance of a physical server."""
        self.mocked_api().dcim.devices.get.return_value = netbox_host
        self.mocked_api().virtualization.virtual_machines.get.return_value = None
        assert isinstance(self.netbox.get_server("physical"), NetboxServer)

    def test_get_server_virtual(self, netbox_virtual_machine):
        """It should return the NetboxServer instance of a virtual machine."""
        self.mocked_api().dcim.devices.get.return_value = None
        self.mocked_api().virtualization.virtual_machines.get.return_value = netbox_virtual_machine
        assert isinstance(self.netbox.get_server("virtual"), NetboxServer)


class TestNetboxServer:
    """Test class for the NetboxServer class."""

    @mock.patch("pynetbox.api")
    def setup_method(self, _, mocked_api):
        """Instantiate the test instance mock."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_api = mocked_api
        self.mgmt_ip = mock.Mock()
        self.mgmt_ip.dns_name = "physical.mgmt.local"
        self.mocked_api.ipam.ip_addresses.get.return_value = self.mgmt_ip

    @pytest.fixture(autouse=True)
    def setup_method_fixture(self, netbox_host, netbox_virtual_machine):
        """Instantiate the test instance with fixtures."""
        # pylint: disable=attribute-defined-outside-init
        self.netbox_host = netbox_host
        self.netbox_virtual_machine = netbox_virtual_machine
        self.physical_server = NetboxServer(api=self.mocked_api, server=self.netbox_host, dry_run=False)
        self.virtual_server = NetboxServer(api=self.mocked_api, server=self.netbox_virtual_machine, dry_run=False)

    def test_init_not_a_server(self):
        """It should raise a NetboxError if the device doesn't have a server role."""
        self.netbox_host.device_role.slug = "not-server"
        with pytest.raises(NetboxError, match="has invalid role not-server"):
            NetboxServer(api=self.mocked_api, server=self.netbox_host)

    def test_virtual_getter(self):
        """It should return a boolean if the device is a virtual machine or a physical host."""
        assert not self.physical_server.virtual
        assert self.virtual_server.virtual

    def test_status_getter(self):
        """It should return the value of the current status of the server, as a string."""
        assert isinstance(self.physical_server.status, str)
        assert self.physical_server.status == "active"

    def test_status_setter_ok(self):
        """It should set the status of the server to its new value, if it's an allowed transition."""
        self.physical_server.status = "failed"
        assert self.netbox_host.save.called_once_with()

    def test_status_setter_virtual(self):
        """It should raise a NetboxError if trying to set the status on a virtual machine."""
        with pytest.raises(
            NetboxError,
            match="Server virtual is a virtual machine, its Netbox status is automatically synced from Ganeti",
        ):
            self.virtual_server.status = "failed"

    def test_status_setter_invalid_transition(self):
        """It should raise a NetboxError if trying to make an invalid status transition."""
        with pytest.raises(
            NetboxError, match="Forbidden Netbox status transition between active and planned for device physical"
        ):
            self.physical_server.status = "planned"

    def test_status_setter_dry_run(self):
        """It should skip setting the status of the server to its new value in dry-run mode."""
        physical_server = NetboxServer(api=self.mocked_api, server=self.netbox_host, dry_run=True)
        physical_server.status = "failed"
        self.netbox_host.save.assert_not_called()
        assert self.physical_server.status == "active"

    def test_fqdn_getter_ok(self):
        """It should return the FQDN of the device from its primary IP."""
        assert self.physical_server.fqdn == "physical.example.com"

    def test_fqdn_getter_v6_only(self):
        """It should return the FQDN of the device from its primary IPv6 if the IPv4 doesn't have a DNS name."""
        self.netbox_host.primary_ip4.dns_name = ""
        assert self.physical_server.fqdn == "physical.example.com"

    def test_fqdn_getter_no_dns(self):
        """It should raise a NetboxError if there is no DNS name defined for any primary IP of the device."""
        self.netbox_host.primary_ip4 = None
        self.netbox_host.primary_ip6.dns_name = None
        with pytest.raises(NetboxError, match="Server physical does not have any primary IP with a DNS name set"):
            self.physical_server.fqdn  # pylint: disable=pointless-statement

    def test_mgmt_fqdn_getter(self):
        """It should return the management FQDN of the device and cache it."""
        assert self.physical_server.mgmt_fqdn == "physical.mgmt.local"
        assert self.physical_server.mgmt_fqdn == "physical.mgmt.local"
        self.mocked_api.ipam.ip_addresses.get.assert_called_once_with(device="physical", interface="mgmt")

    def test_mgmt_fqdn_getter_no_mgmt(self):
        """It should raise a NetboxError if the management FQDN is not set."""
        self.mgmt_ip.dns_name = ""
        with pytest.raises(NetboxError, match="Server physical has no management interface with a DNS name set"):
            self.physical_server.mgmt_fqdn  # pylint: disable=pointless-statement

    def test_mgmt_fqdn_getter_virtual(self):
        """It should raise a NetboxError if trying to get the management FQDN on a virtual machine."""
        with pytest.raises(
            NetboxError, match="Server virtual is a virtual machine, does not have a management address"
        ):
            self.virtual_server.mgmt_fqdn  # pylint: disable=pointless-statement

    def test_asset_tag_fqdn_getter(self):
        """It should return the management FQDN of the asset tag of the device."""
        assert self.physical_server.asset_tag_fqdn == "asset1234.mgmt.local"

    def test_as_dict_physical(self):
        """It should return the dictionary representation of the physical server."""
        as_dict = self.physical_server.as_dict()
        assert as_dict["cluster"] is None
        assert not as_dict["is_virtual"]
        assert as_dict["name"] == "physical"

    def test_as_dict_virtual(self):
        """It should return the dictionary representation of the virtual machine."""
        as_dict = self.virtual_server.as_dict()
        assert as_dict["cluster"] == {"id": 1, "name": "testcluster"}
        assert as_dict["is_virtual"]
        assert as_dict["name"] == "virtual"



"""Redfish module."""
import ipaddress
import json
import logging
import re
import time
from abc import abstractmethod
from datetime import datetime, timedelta
from enum import Enum
from io import BufferedReader
from pathlib import Path
from typing import Any, Optional, Union

import urllib3
from packaging import version
from requests import Response
from requests.exceptions import RequestException
from wmflib.requests import http_session

from spicerack.decorators import retry
from spicerack.exceptions import SpicerackError

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
logger = logging.getLogger(__name__)


class RedfishError(SpicerackError):
    """General errors raised by this module."""


class RedfishTaskNotCompletedError(RedfishError):
    """Raised when a Redfish task is not found on the server."""


class ChassisResetPolicy(Enum):
    """Subset of available Chassis.Reset policies compatible with all supported vendors (at this moment only Dell)."""

    FORCE_OFF: str = "ForceOff"
    """Turn off the unit immediately (nongraceful shutdown)."""
    FORCE_RESTART: str = "ForceRestart"
    """Shut down immediately and nongracefully and restart the system."""
    GRACEFUL_RESTART: str = "GracefulRestart"
    """Shut down gracefully and power on."""
    GRACEFUL_SHUTDOWN: str = "GracefulShutdown"
    """Shut down gracefully and power off."""
    ON: str = "On"
    """Turn on the unit."""


class DellSCPRebootPolicy(Enum):
    """Available Dell SCP (Server Configuration Profiles) reboot policies."""

    FORCED: str = "Forced"
    """Issue an immediate hard reboot without notifying the operating system."""
    GRACEFUL: str = "Graceful"
    """Issue a reboot notifying the operating system."""
    NO_REBOOT: str = "NoReboot"
    """Do not reboot right now, the Redfish task will be pending the next reboot to apply the changes."""


class DellSCPPowerStatePolicy(Enum):
    """Available Dell SCP (Server Configuration Profiles) final power state after an operation policies."""

    OFF: str = "Off"
    """Keep the host powered off after the operation."""
    ON: str = "On"
    """Turn the host power back on after the operation."""


class DellSCPTargetPolicy(Enum):
    """Available sections of the Dell SCP (Server Configuration Profiles) to target."""

    ALL = "ALL"
    """All settings."""
    BIOS = "BIOS"
    """Only BIOS settings."""
    IDRAC = "IDRAC"
    """Only iDRAC settings."""
    NIC = "NIC"
    """Only network interfaces settings."""
    RAID = "RAID"
    """Only RAID controller settings."""


class Redfish:
    """Manage Redfish operations on a specific device."""

    def __init__(
        self,
        hostname: str,
        interface: Union[ipaddress.IPv4Interface, ipaddress.IPv6Interface],
        username: str,
        password: str,
        *,
        dry_run: bool = True,
    ):
        """Initialize the instance.

        Arguments:
            hostname: the hostname (not FQDN) the management console belongs to.
            interface: the interface of the management console to connect to.
            username: the API username.
            password: the API password.
            dry_run: whether this is a DRY-RUN.

        """
        self._dry_run = dry_run
        self._hostname = hostname
        self._interface = interface
        self._username = username
        self._password = password

        self._http_session = http_session(".".join((self.__module__, self.__class__.__name__)), timeout=10)
        # TODO: evaluate if we should create an intermediate CA for managament consoles
        self._http_session.verify = False  # The devices have a self-signed certificate
        self._http_session.auth = (self._username, self._password)
        self._http_session.headers.update({"Accept": "application/json"})

        self._upload_session = http_session(".".join((self.__module__, self.__class__.__name__)), timeout=60 * 30)
        self._upload_session.verify = False  # The devices have a self-signed certificate
        self._upload_session.auth = (self._username, self._password)
        self._upload_session.headers.update({"Accept": "application/json"})

        self._oob_info: dict = {}
        self._system_info: dict = {}
        self._updateservice_info: dict = {}

    def __str__(self) -> str:
        """String representation of the instance."""
        return f"{self._username}@{self._hostname} ({self._interface.ip})"

    @property
    @abstractmethod
    def system_manager(self) -> str:
        """Property to return the System manager."""

    @property
    @abstractmethod
    def oob_manager(self) -> str:
        """Property to return the Out of Band manager."""

    def _update_system_info(self) -> None:
        """Property to return a dict of manager metadata."""
        self._system_info = self.request("get", self.system_manager).json()

    def _update_oob_info(self) -> None:
        """Update the data Out of Band manager info."""
        self._oob_info = self.request("get", self.oob_manager).json()

    @property
    def update_service(self) -> str:
        """Property to return the Out of Band manager."""
        # for now this is the same for both dell and supermicro
        return "/redfish/v1/UpdateService"

    @property
    @abstractmethod
    def log_entries(self) -> str:
        """Property to return the uri for the log entries."""

    @property
    @abstractmethod
    def reboot_message_id(self) -> str:
        """Property to return the Message Id for reboot log entries."""

    @property
    def hostname(self) -> str:
        """Getter for the device hostname."""
        return self._hostname

    @property
    def interface(self) -> Union[ipaddress.IPv4Interface, ipaddress.IPv6Interface]:
        """Getter for the management interface address with netmask."""
        return self._interface

    @property
    def system_info(self) -> dict:
        """Property to return the system info as a dict."""
        if not self._system_info:
            self._update_system_info()
        return self._system_info

    @property
    def oob_info(self) -> dict:
        """Property to return the oob info as a dict."""
        if not self._oob_info:
            self._update_oob_info()
        return self._oob_info

    @property
    def updateservice_info(self) -> dict:
        """Property to return a dict of manager metadata."""
        if not self._updateservice_info:
            result = self.request("get", self.update_service)
            self._updateservice_info = result.json()
        return self._updateservice_info

    @property
    def oob_model(self) -> str:
        """Property to return a string representing the model."""
        return self.oob_info["Model"]

    @property
    def firmware_version(self) -> version.Version:
        """Property to return a version instance representing the firmware version."""
        self._update_oob_info()
        return version.parse(self.oob_info["FirmwareVersion"])

    @property
    def bios_version(self) -> version.Version:
        """Property to return a instance representing the Bios version."""
        self._update_system_info()
        return version.parse(self.system_info["BiosVersion"])

    @property
    def model(self) -> str:
        """Property to return a string representing the model."""
        return self.system_info["Model"]

    @property
    def manufacturer(self) -> str:
        """Property to return a string representing the model."""
        return self.system_info["Manufacturer"]

    @property
    def pushuri(self) -> str:
        """Property representing the HttpPushUri of the idrac for uploading firmwares to it."""
        try:
            return self.updateservice_info["HttpPushUri"]
        except KeyError:
            raise NotImplementedError from KeyError

    @property
    def multipushuri(self) -> str:
        """Property representing the MultipartHttpPushUri of the idrac for uploading firmwares to it."""
        return self.updateservice_info["MultipartHttpPushUri"]

    def upload_file(self, file_path: Path, reboot: bool = False) -> str:
        """Upload a file to the firmware directory via redfish and return the job_id URI.

        Arguments:
            file_path: The file path to upload.
            reboot: if true immediately reboot the server.

        """
        with file_path.open("rb") as file_handle:
            return self.multipush_upload(file_handle, file_path.name, reboot)

    def multipush_upload(self, file_handle: BufferedReader, filename: str, reboot: bool = False) -> str:
        """Upload a file via redfish and return its job_id URI.

        Arguments:
            file_handle: On open file handle to the object to upload.
            filename: filename name to use for upload.
            reboot: if true immediately reboot the server.

        """
        operation = "Immediate" if reboot else "OnReset"
        payload = {"Targets": [], "@Redfish.OperationApplyTime": operation, "Oem": {}}
        files = {
            "UpdateParameters": (None, json.dumps(payload), "application/json"),
            "UpdateFile": (filename, file_handle, "application/octet-stream"),
        }
        job_id = self.submit_files(files)
        logger.debug("upload has task ID: %s", job_id)
        return job_id

    @staticmethod
    def most_recent_member(members: list[dict], key: str) -> dict:
        """Return the most recent member of members result from dell api.

        Members will be sorted on key and the most recent value is returned.
        The value of key is assumed to be an iso date.

        Arguments:
            members: A list of dicts returned from the dell api.
            key: The key to search on.

        """

        def sorter(element: dict) -> datetime:
            """Sort by datetime."""
            return datetime.fromisoformat(element[key])

        return sorted(members, key=sorter)[-1]

    def last_reboot(self) -> datetime:
        """Get the the last reboot time."""
        # TODO: we can possibly use filter once all OOB's are updated.  e.g.
        # Lclog/Entries?$filter=MessageId eq 'reboot_code'
        # currently we get the following on some older models
        # Message=Querying is not supported by the implementation, MessageArgs=$filter"
        last_reboot = datetime.fromisoformat("1970-01-01T00:00:00-00:00")
        results = self.request("get", self.log_entries).json()
        # use ends with as sometimes there is an additional string prefix to the code e.g. IDRAC.2.7.RAC0182
        members = [m for m in results["Members"] if m["MessageId"].endswith(self.reboot_message_id)]
        if members:
            last_reboot = datetime.fromisoformat(self.most_recent_member(members, "Created")["Created"])
        logger.debug("%s: last reboot %s", self._hostname, last_reboot)
        return last_reboot

    @retry(
        tries=240,
        delay=timedelta(seconds=10),
        backoff_mode="constant",
        exceptions=(RedfishError,),
    )
    def wait_reboot_since(self, since: datetime) -> None:
        """Wait for idrac/redfish to become responsive.

        Arguments:
            since: The datetime of the last reboot.

        """
        self.check_connection()
        latest = self.last_reboot()
        if since >= latest:
            raise RedfishError("no new reboot detected")
        logger.debug("%s: new management console reboot detected %s", self._hostname, latest)

    def request(self, method: str, uri: str, **kwargs: Any) -> Response:
        """Perform a request against the target Redfish instance with the provided HTTP method and data.

        Arguments:
            uri: the relative URI to request.
            method: the HTTP method to use (e.g. "post").
            **kwargs: arbitrary keyword arguments, to be passed requests.

        Raises:
            spicerack.redfish.RedfishError: if the response status code is between 400 and 600 or if the given uri
            does not start with a slash (/) or if the request couldn't be performed.

        """
        if uri[0] != "/":
            raise RedfishError(f"Invalid uri {uri}, it must start with a /")

        url = f"https://{self._interface.ip}{uri}"

        if self._dry_run and method.lower() not in ("head", "get"):  # RW call
            logger.info("Would have called %s on %s", method, url)
            return self._get_dummy_response()

        try:
            response = self._http_session.request(method, url, **kwargs)
        except RequestException as e:
            message = f"Failed to perform {method.upper()} request to {url}"
            if self._dry_run:
                logger.error("%s: %s", message, e)
                return self._get_dummy_response()

            raise RedfishError(message) from e

        if not response.ok:
            raise RedfishError(
                f"{method.upper()} {url} returned HTTP {response.status_code} with message:\n{response.text}"
            )

        return response

    def _parse_submit_task(self, response: Response) -> str:
        """Submit a request that generates a task, return the URI of the submitted task.

        Arguments:
            response: the response to parse.

        Raises:
            spicerack.redfish.RedfishError: if the response status code is not 202 or there is no Location header.

        """
        uri = response.request.path_url
        if response.status_code != 202:
            raise RedfishError(
                f"Unable to start task for {uri}, expected HTTP 202, "
                f"got HTTP {response.status_code} instead:\n{response.text}"
            )

        # requests allow to access headers with any capitalization
        if "Location" not in response.headers or not response.headers["Location"]:
            raise RedfishError(
                "Unable to get the task URI to poll results for the {uri} request. Returned headers:\n"
                f"{response.headers}"
            )

        return response.headers["Location"]

    def submit_files(self, files: dict) -> str:
        """Submit a upload file request that generates a task, return the URI of the submitted task.

        Arguments:
            uri: the relative URI to request.
            files: the files to upload to send in the request.

        """
        if self._dry_run:
            return "/"

        # BUG: timeout is not honoured by self.request
        # response = self.request('post', self.multipushuri, files=files)
        response = self._upload_session.post(f"https://{self.interface.ip}{self.multipushuri}", files=files)
        return self._parse_submit_task(response)

    def submit_task(self, uri: str, data: Optional[dict] = None, method: str = "post") -> str:
        """Submit a request that generates a task, return the URI of the submitted task.

        Arguments:
            uri: the relative URI to request.
            data: the data to send in the request.
            method: the HTTP method to use, if not the default one.

        """
        if self._dry_run:
            return "/"

        response = self.request(method, uri, json=data)
        return self._parse_submit_task(response)

    def check_connection(self) -> None:
        """Check the connection with the Redfish API.

        Raises:
            spicerack.redfish.RedfishError: if unable to connect to Redfish API.

        """
        logger.info("Testing Redfish API connection to %s (%s)", self._hostname, self._interface.ip)
        self.request("get", "/redfish")

    @retry(
        tries=30,
        delay=timedelta(seconds=30),
        backoff_mode="constant",
        exceptions=(RedfishTaskNotCompletedError,),
        failure_message="Polling task",
    )
    def poll_task(self, uri: str) -> dict:
        """Poll a Redfish task until the results are available and return them.

        Arguments:
           uri: the URI of the task, usually returned as Location header by the originating API call.

        Raises:
            spicerack.redfish.RedfishError: if the response from the server is outside the expected values of HTTP
            200 or 202.
            spicearck.redfish.RedfishTaskNotCompletedError: if the task is not yet completed.

        """
        if self._dry_run:
            return {}

        response = self.request("get", uri)
        if response.status_code not in (200, 202):
            raise RedfishError(f"{uri} returned HTTP {response.status_code}:\n{response.text}")

        results = response.json()
        if response.status_code == 200 and "@odata.id" not in results:  # Task completed, got data without metadata
            return results

        for message in results["Messages"]:
            if "Oem" in message:
                continue  # Skip Oem messages, they might have any custom structure

            # Some older Dell implementation use both keys in the same API response :/
            message_id = message.get("MessageId", message.get("MessageID", "N.A."))
            logger.info("[%s] %s", message_id, message["Message"])

        try:
            end_time = datetime.fromisoformat(results.get("EndTime", "1970-01-01T00:00:00")).timestamp()
        except ValueError:
            # Endtime is TIME_NA
            end_time = 0

        if end_time == 0:
            raise RedfishTaskNotCompletedError(
                f"{results['Id']} not completed yet: status={results['TaskStatus']}, state={results['TaskState']}, "
                f"completed={results.get('PercentComplete', 'unknown')}%"
            )

        return results  # When a task is polled after returning the data, will return again the metadata

    def find_account(self, username: str) -> tuple[str, str]:
        """Find the account for the given username and return its URI.

        Arguments:
            username: the username to search for.

        Returns:
            A 2-element tuple with the URI for the account and the ETag header value of the GET response.

        Raises:
            spicerack.redfish.RedfishError: if unable to find the account.

        """
        accounts = self.request("get", f"{self.oob_manager}/Accounts").json()
        uris = [account["@odata.id"] for account in accounts["Members"]]
        for uri in uris:
            response = self.request("get", uri)
            if response.json()["UserName"] == username:
                return uri, response.headers.get("ETag", "")

        raise RedfishError(f"Unable to find account for username {username}")

    def change_user_password(self, username: str, password: str) -> None:
        """Change the password for the account with the given username.

        If the username matches the one used by the instance to connect to the API, automatically updates the instance
        value so that the instance will keep working.

        Arguments:
            username: the username to search for.
            password: the new password to set.

        Raises:
            spicerack.redfish.RedfishError: if unable to find the user or update its password.

        """
        user_uri, etag = self.find_account(username)
        logger.info("Changing password for the account with username %s: %s", username, user_uri)
        response = self.request(
            "patch", user_uri, json={"UserName": username, "Password": password}, headers={"If-Match": etag}
        )
        if response.status_code != 200:
            raise RedfishError(f"Got unexpected HTTP {response.status_code}, expected 200:\n{response.text}")

        if self._username == username and not self._dry_run:
            self._password = password
            self._http_session.auth = (self._username, self._password)
            logger.info("Updated current instance password to the new password")

        for message in response.json().get("@Message.ExtendedInfo", []):
            identifier = " ".join((message.get("MessageId", ""), message.get("Severity", ""))).strip()
            logger.info(
                "[%s] %s | Resolution: %s", identifier, message.get("Message", ""), message.get("Resolution", "")
            )

    def get_power_state(self) -> str:
        """Return the current power state of the device."""
        response = self.request("get", "/redfish/v1/Chassis/System.Embedded.1").json()
        return response["PowerState"]

    def chassis_reset(self, action: ChassisResetPolicy) -> None:
        """Perform a reset of the chassis power status.

        Arguments:
            action: the reset policy to use.

        Raises:
            spicerack.redfish.RedfishError: if unable to perform the reset.

        """
        logger.info("Resetting chassis power status for %s to %s", self._hostname, action.value)
        response = self.request(
            "post",
            "/redfish/v1/Systems/System.Embedded.1/Actions/ComputerSystem.Reset",
            json={"ResetType": action.value},
        )
        if response.status_code != 204 and not self._dry_run:
            raise RedfishError(
                f"Got unexpected response HTTP {response.status_code}, expected HTTP 204: {response.text}"
            )

    @staticmethod
    def _get_dummy_response() -> Response:
        """Return a dummy requests's Response to be used in dry-run mode."""
        response = Response()
        response.status_code = 200
        return response


class DellSCP:
    """Reprenset a Dell System Configuration Profile configuration as returned by Redfish API."""

    def __init__(self, config: dict, target: DellSCPTargetPolicy, *, allow_new_attributes: bool = False):
        """Parse the Redfish API response.

        Arguments:
            config: the configuration as returned by Redfish API.
            target: describe which sections of the configuration are represented in the loaded configuration.
            allow_new_attributes: when set to :py:data:`True` it allows the creation of new attributes not
                already present in the provided configuration that otherwise would raise an exception. This is
                useful for example when changing the boot mode between Uefi and Bios that changes the keys present.

        """
        self._config = config
        self._target = target
        self._allow_new_attributes = allow_new_attributes
        # Track if the Components property have been emptied, allowing the creation of new components
        self._emptied_components = False

    @property
    def config(self) -> dict:
        """Getter for the whole configuration in Dell's format."""
        return self._config

    @property
    def target(self) -> DellSCPTargetPolicy:
        """Getter for the target that the current configuration represents."""
        return self._target

    @property
    def service_tag(self) -> str:
        """Getter for the device Dell's Service Tag."""
        return self._config["SystemConfiguration"]["ServiceTag"]

    @property
    def model(self) -> str:
        """Getter for the device Dell's model."""
        return self._config["SystemConfiguration"]["Model"]

    @property
    def timestamp(self) -> datetime:
        """Getter for the timestamp when the configuration dump was generated."""
        return datetime.strptime(self._config["SystemConfiguration"]["TimeStamp"], "%c")

    @property
    def comments(self) -> list[str]:
        """Getter for the comments associated with the configuration."""
        return [
            comment["Comment"]
            for comment in self._config["SystemConfiguration"].get("Comments", [])
            if "Comment" in comment
        ]

    @property
    def components(self) -> dict[str, dict[str, str]]:
        """Getter for the components present in the configuration in a simplified view.

        The returned dictionary where all the keys are recursively sorted and has the following format::

            {
                '<component name>': {
                    'key1': 'value1',
                    'key2': 'value2',
                },
            }

        """
        components: dict[str, dict[str, str]] = {}
        for component in self._config["SystemConfiguration"].get("Components", []):
            components[component["FQDD"]] = {}
            for attribute in component.get("Attributes", []):
                components[component["FQDD"]][attribute["Name"]] = attribute["Value"]

        # Sort the components recursively
        return {component: dict(sorted(components[component].items())) for component in sorted(components)}

    def set(self, component_name: str, attribute_name: str, attribute_value: str) -> bool:
        """Update the current configuration setting to the new value for the given key in the given component.

        Notes:
            This updates only the instance representation of the instance. To push and apply the changes to the server
            see :py:meth:`spicerack.redfish.RedfishDell.scp_push`.
            In order to add attributes not present the instance must have been created with `allow_new_attributes` set
            to :py:data:`True` or the :py:meth:`spicerack.redfish.DellSCP.empty_component` method called. This last one
            allows to automatically create any missing component while setting attributes.

        Arguments:
            component_name: the name of the component the settings belongs to.
            attribute_name: the attribute name whose value needs to be updated.
            attribute_value: the new value for the attribute to set.

        Returns:
            :py:data:`True` if the value was added or changed, :py:data:`False` if it had already the correct value.

        Raises:
            spicerack.redfish.RedfishError: if unable to find the given component or attribute and the creation of new
            items is not allowed.

        """

        def new_attribute() -> dict[str, str]:
            """Local helper that returns a new attribute to append to the component."""
            attribute = {
                "Name": attribute_name,
                "Value": attribute_value,
                "Set On Import": "True",
                "Comment": "Read and Write",
            }
            logger.info(
                "Created attribute %s -> %s (with Set On Import True) with value %s",
                component_name,
                attribute_name,
                attribute_value,
            )
            return attribute

        for component in self._config["SystemConfiguration"]["Components"]:
            if component["FQDD"] != component_name:
                continue

            for i in range(len(component["Attributes"])):
                attribute = component["Attributes"][i]
                if attribute["Name"] != attribute_name:
                    continue

                # Attribute found, update it if different, consider comma-separated lists identical with both ',' and
                # ', ' as separators.
                if attribute["Value"].replace(", ", ",") == attribute_value.replace(", ", ","):
                    logger.info(
                        "Skipped set of attribute %s -> %s, has already the correct value: %s",
                        component_name,
                        attribute_name,
                        attribute["Value"],
                    )
                    return False

                logger.info(
                    "Updated value for attribute %s -> %s%s: %s => %s",
                    component_name,
                    attribute_name,
                    " (marked Set On Import to True)" if attribute["Set On Import"] == "False" else "",
                    attribute["Value"],
                    attribute_value,
                )
                attribute["Value"] = attribute_value
                attribute["Set On Import"] = "True"

                return True

            # Attribute not found, add it or raise
            if self._allow_new_attributes or self._emptied_components:
                component["Attributes"].append(new_attribute())
                return True

            raise RedfishError(f"Unable to find attribute {component_name} -> {attribute_name}")

        # Component not found, add it or raise
        if self._emptied_components:
            self._config["SystemConfiguration"]["Components"].append(
                {"FQDD": component_name, "Attributes": [new_attribute()]}
            )
            return True

        raise RedfishError(f"Unable to find component {component_name}")

    def update(self, changes: dict[str, dict[str, str]]) -> bool:
        """Bulk update the current configuration with the set of changes provided.

        Notes:
            This updates only the instance representation of the instance. To push and apply the changes to the server
            see :py:meth:`spicerack.redfish.RedfishDell.scp_push`.

        Arguments:
            changes: a dictionary of changes to apply in the same format of the one returned by
                :py:meth:`spicerack.redfish.DellSCP.components`.

        Returns:
            :py:data:`True` if any of the values produced a change, :py:data:`False` if no change was made.

        Raises:
            spicerack.redfish.RedfishError: if unable to apply all the changes.

        """
        was_changed = False
        for component, attributes in changes.items():
            for name, value in attributes.items():
                was_changed |= self.set(component, name, value)

        return was_changed

    def empty_components(self) -> None:
        """Empty the current Components from the configuration, allowing to create a new configuration from scratch.

        After calling this method is possible to set values for non-existing components that would otherwise raise an
        exception.

        """
        self._config["SystemConfiguration"]["Components"] = []
        self._emptied_components = True


class RedfishSupermicro(Redfish):
    """Redfish class for SuperMicro servers."""

    @property
    def system_manager(self) -> str:
        """Property to return the System manager."""
        return "/redfish/v1/Systems/1"

    @property
    def oob_manager(self) -> str:
        """String representing the Out of Band manager key."""
        return "/redfish/v1/Managers/1"

    @property
    def log_entries(self) -> str:
        """String representing the log entries uri."""
        return "/redfish/v1/Managers/1/LogServices/Log1/Entries"

    @property
    def reboot_message_id(self) -> str:
        """String representing the message Id of the reboot."""
        return "Event.1.0.SystemPowerAction"


class RedfishDell(Redfish):
    """Dell specific Redfish support."""

    scp_base_uri: str = "/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Oem/EID_674_Manager"
    """The Dell's SCP push base URI."""

    def __init__(
        self,
        hostname: str,
        interface: Union[ipaddress.IPv4Interface, ipaddress.IPv6Interface],
        username: str,
        password: str,
        *,
        dry_run: bool = True,
    ):
        """Override parent's constructor."""
        super().__init__(hostname, interface, username, password, dry_run=dry_run)
        self._generation = 0

    @property
    def system_manager(self) -> str:
        """Property to return the System manager."""
        return "/redfish/v1/Systems/System.Embedded.1"

    @property
    def oob_manager(self) -> str:
        """String representing the Out of Band manager key."""
        return "/redfish/v1/Managers/iDRAC.Embedded.1"

    @property
    def log_entries(self) -> str:
        """String representing the log entries uri."""
        if self.firmware_version < version.Version("4.10"):
            return "/redfish/v1/Managers/Logs/Lclog"
        return "/redfish/v1/Managers/iDRAC.Embedded.1/LogServices/Lclog/Entries"

    @property
    def reboot_message_id(self) -> str:
        """String representing the message Id of the reboot."""
        return "RAC0182"

    @property
    def generation(self) -> int:
        """Property representing the generation of the idrac.

        This is often 13 for idrac8 and 14 for idrac9.  This property allows us to add workarounds
        for older idrac models
        """
        if self._generation == 0:
            match = re.search(r"\d+", self.oob_model)
            if match is None:
                logger.error("%s: Unrecognized iDRAC model %s, setting generation to 1", self._hostname, self.oob_model)
                # Setting this to one allows use to continue but assumes the minimal level of support
                self._generation = 1
            else:
                self._generation = int(match.group(0))
        logger.debug("%s: iDRAC generation %s", self._hostname, self._generation)
        return self._generation

    @retry(
        tries=240,
        delay=timedelta(seconds=10),
        backoff_mode="constant",
        exceptions=(RedfishError,),
    )
    def wait_reboot_since(self, since: datetime) -> None:
        """Wait for idrac/redfish to become responsive.

        Arguments:
            since: the datetime of the last reboot.

        """
        self.check_connection()
        if self._generation < 14:
            # Probing the Gen13/iDRAC8 devices too early seems to cause the redfish deamon to crash
            print("sleeping for 2 mins to let idrac boot")
            time.sleep(120)
        latest = self.last_reboot()
        if since >= latest:
            raise RedfishError("no new reboot detected")
        logger.debug("%s: new management console reboot detected %s", self._hostname, latest)
        # Its still takes a bit of time for redfish to fully come only so
        # We just arbitrarily sleep for a bit
        sleep_secs = 30
        logger.debug("%s: sleeping for %d secs", self._hostname, sleep_secs)
        time.sleep(sleep_secs)

    def scp_dump(
        self, target: DellSCPTargetPolicy = DellSCPTargetPolicy.ALL, *, allow_new_attributes: bool = False
    ) -> DellSCP:
        """Dump and return the SCP (Server Configuration Profiles) configuration.

        Arguments:
            target: choose which sections to dump.
            allow_new_attributes: when set to :py:data:`True` it allows the creation of new attributes not
                already present in the retrieved configuration that otherwise would raise an exception. This is
                useful for example when changing the boot mode between Uefi and Bios that changes the keys present.

        Raises:
            spicerack.redfish.RedfishError: if the API call fail.
            spicerack.redfish.RedfishTaskNotCompletedError: if unable to fetch the dumped results.

        """
        data = {"ExportFormat": "JSON", "ShareParameters": {"Target": target.value}}
        task_uri = self.submit_task(f"{self.scp_base_uri}.ExportSystemConfiguration", data)

        return DellSCP(self.poll_task(task_uri), target, allow_new_attributes=allow_new_attributes)

    def scp_push(
        self,
        scp: DellSCP,
        *,
        reboot: DellSCPRebootPolicy = DellSCPRebootPolicy.NO_REBOOT,
        power_state: DellSCPPowerStatePolicy = DellSCPPowerStatePolicy.ON,
        preview: bool = True,
    ) -> dict:
        """Push the SCP (Server Configuration Profiles) configuration.

        Arguments:
            scp: the configuration that will pushed to the server.
            reboot: which reboot policy to use to apply the changes.
            power_state: which final power state policy to use to apply to the host after the changes have been
                applied.
            preview: if :py:data:`True` perform a test push of the SCP data. This will tell if the file
                parses correctly and would not result in any writes. The comments will tell if the new configuration
                would not produce any changes. Forces the reboot parameter to be
                :py:const:`spicerack.redfish.DellSCPRebootPolicy.NO_REBOOT`.

        Returns:
            The results of the push operation.

        Raises:
            spicerack.redfish.RedfishError: if the API call fail.
            spicerack.redfish.RedfishTaskNotCompletedError: if unable to fetch the dumped results.

        """
        if preview:
            uri = "ImportSystemConfigurationPreview"
            reboot = DellSCPRebootPolicy.NO_REBOOT
        else:
            uri = "ImportSystemConfiguration"

        data = {
            "ImportBuffer": json.dumps(scp.config),  # The API requires a JSON-encoded string inside a JSON payload.
            "ShareParameters": {"Target": scp.target.value},
            "HostPowerState": power_state.value,
            "ShutdownType": reboot.value,
        }

        task_id = self.submit_task(f"{self.scp_base_uri}.{uri}", data)

        return self.poll_task(task_id)
